import copy
from metadrive.component.sensors.vehicle_panel import VehiclePanel
import logging
from typing import Dict, Any

from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.road_network import Road
from metadrive.constants import TerminationState
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.manager.spawn_manager import SpawnManager
from metadrive.utils import setup_logger, get_np_random, Config

MULTI_AGENT_METADRIVE_DEFAULT_CONFIG = dict(
    # ===== Multi-agent =====
    is_multi_agent=True,
    num_agents=15,
    # If num_agents is set to None, then endless vehicles will be added only the empty spawn points exist
    random_agent_model=False,
    # If True, the spawn position will be deterministic for each episode, usually used for deterministic test
    force_seed_spawn_manager=False,
    # Whether to terminate a vehicle if it crash with others. Since in MA env the crash is extremely dense, so
    # frequently done might not be a good idea.
    crash_done=True,
    out_of_road_done=True,
    delay_done=25,  # Put the dead vehicles in place for 5 seconds before removing them.
    # Whether the vehicle can rejoin the episode
    allow_respawn=True,
    spawn_roads=[Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3)],
    # The maximum length of the episode. If allow respawn, then this is the maximum step that respawn can happen. After
    # that, the episode won't terminate until all existing vehicles reach their horizon or done. The vehicle specified
    # horizon is also this value.
    horizon=1000,
    max_step_per_agent=1000,  # Per agent maximum episode steps
    # Use to determine what neighborhood means
    neighbours_distance=10,
    # ===== Vehicle Setting =====
    vehicle_config=dict(
        lidar=dict(num_lasers=72, distance=40, num_others=0),
        random_color=True,
        spawn_lane_index=(FirstPGBlock.NODE_1, FirstPGBlock.NODE_2, 0),
        _specified_spawn_lane=False,  # automatically filled
        _specified_destination=False,  # automatically filled
        # We remove dynamics randomization in Multi-agent environments to make the results aligned with previous
        # results published in papers. See
        # https://github.com/metadriverse/metadrive/issues/161#issuecomment-1080114029
        # for more information
        vehicle_model="static_default",
    ),
    target_vehicle_configs=dict(),
    # ===== New Reward Setting =====
    out_of_road_penalty=10,
    crash_vehicle_penalty=10,
    crash_object_penalty=10,
    crash_vehicle_cost=1,
    crash_object_cost=1,
    out_of_road_cost=0,  # Do not count out of road into cost!
    # ===== Environmental Setting =====
    traffic_density=0.0,
    camera_height=4,
    map_file_path="",
    interface_panel=["dashboard"],
)


class MultiAgentMetaDrive(MetaDriveEnv):
    """
    This serve as the base class for Multi-agent MetaDrive!
    """

    # A list of road instances denoting which roads afford spawn points. If not set, then search for all
    # possible roads and spawn new agents in them if possible.

    @staticmethod
    def default_config() -> Config:
        config = MetaDriveEnv.default_config()
        config.update(MULTI_AGENT_METADRIVE_DEFAULT_CONFIG)
        return config

    def _post_process_config(self, config):
        from metadrive.manager.spawn_manager import SpawnManager

        config = super(MultiAgentMetaDrive, self)._post_process_config(config)
        ret_config = config
        # merge basic vehicle config into target vehicle config
        target_vehicle_configs = dict()
        num_agents = (
            ret_config["num_agents"] if ret_config["num_agents"] != -1 else SpawnManager.max_capacity(
                config["spawn_roads"],
                config["map_config"]["exit_length"],
                config["map_config"]["lane_num"],
            )
        )
        for id in range(num_agents):
            agent_id = "agent{}".format(id)
            config = copy.deepcopy(ret_config["vehicle_config"])
            if agent_id in ret_config["target_vehicle_configs"]:
                config["_specified_spawn_lane"] = (
                    True if "spawn_lane_index" in ret_config["target_vehicle_configs"][agent_id] else False
                )
                config["_specified_destination"] = (
                    True if "destination" in ret_config["target_vehicle_configs"][agent_id] else False
                )
                config.update(ret_config["target_vehicle_configs"][agent_id])
            target_vehicle_configs[agent_id] = config
        ret_config["target_vehicle_configs"] = target_vehicle_configs
        if ret_config["use_render"] and ret_config["disable_model_compression"]:
            logging.warning("Turn disable_model_compression=True can decrease the loading time!")

        if "prefer_track_agent" in config and config["prefer_track_agent"]:
            ret_config["target_vehicle_configs"][config["prefer_track_agent"]]["use_special_color"] = True
        ret_config["vehicle_config"]["random_agent_model"] = ret_config["random_agent_model"]
        return ret_config

    def done_function(self, vehicle_id):
        done, done_info = super(MultiAgentMetaDrive, self).done_function(vehicle_id)
        if done_info[TerminationState.CRASH] and (not self.config["crash_done"]):
            assert (
                done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.SUCCESS]
                or done_info[TerminationState.OUT_OF_ROAD]
            )
            if not (done_info[TerminationState.SUCCESS] or done_info[TerminationState.OUT_OF_ROAD]):
                # Does not revert done if high-priority termination happens!
                done = False

        if done_info[TerminationState.OUT_OF_ROAD] and (not self.config["out_of_road_done"]):
            assert (
                done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.SUCCESS]
                or done_info[TerminationState.OUT_OF_ROAD]
            )
            if not done_info[TerminationState.SUCCESS]:
                done = False

        return done, done_info

    def step(self, actions):
        o, r, tm, tc, i = super(MultiAgentMetaDrive, self).step(actions)
        o, r, tm, tc, i = self._after_vehicle_done(o, r, tm, tc, i)

        # Update respawn manager
        if self.episode_step >= self.config["horizon"]:
            self.agent_manager.set_allow_respawn(False)
        new_obs_dict, new_info_dict = self._respawn_vehicles(randomize_position=self.config["random_traffic"])
        if new_obs_dict:
            for new_id, new_obs in new_obs_dict.items():
                o[new_id] = new_obs
                r[new_id] = 0.0
                i[new_id] = new_info_dict[new_id]
                tm[new_id] = False
                tc[new_id] = False

        # Update __all__
        d_all = False
        if (self.config["horizon"] is not None):  # No agent alive or a too long episode happens
            if (self.episode_step >= self.config["horizon"] and (all(tc.values()) or all(tm.values()))
                    or (self.episode_step >= 5 * self.config["horizon"])):
                d_all = True
        if len(self.vehicles) == 0:  # No agent alive
            d_all = True
        tm["__all__"] = d_all
        if tm["__all__"]:
            for k in tm.keys():
                tm[k] = True

        return o, r, tm, tc, i

    def _after_vehicle_done(
        self, obs: Dict[str, Any], reward: Dict[str, float], terminated: Dict[str, bool], truncated: Dict[str, bool],
        info: Dict[str, Any]
    ):
        for v_id, v_info in info.items():
            if v_info.get("episode_length", 0) >= self.config["horizon"]:
                if terminated[v_id] is not None:
                    info[v_id][TerminationState.MAX_STEP] = True
                    terminated[v_id] = True
                    self.dones[v_id] = True
        for dead_vehicle_id, termed in terminated.items():
            if termed:
                self.agent_manager.finish(
                    dead_vehicle_id,
                    ignore_delay_done=info[dead_vehicle_id].get(TerminationState.SUCCESS, False),
                )
                self._update_camera_after_finish()
        return obs, reward, terminated, truncated, info

    def _update_camera_after_finish(self):
        if (self.main_camera is not None
                and self.current_track_vehicle.id not in self.engine.agent_manager._active_objects
                and self.engine.task_manager.hasTaskNamed(self.main_camera.CHASE_TASK_NAME)):
            self.switch_to_third_person_view()

    def _get_observations(self):
        return {name: self.get_single_observation() for name in self.config["target_vehicle_configs"].keys()}

    def _respawn_vehicles(self, randomize_position=False):
        new_obs_dict = {}
        new_info_dict = {}
        if not self.agent_manager.allow_respawn:
            return new_obs_dict, new_info_dict
        while True:
            new_id, new_obs, step_info = self._respawn_single_vehicle(randomize_position=randomize_position)
            if new_obs is not None:
                new_obs_dict[new_id] = new_obs
                new_info_dict[new_id] = step_info
            else:
                break
        return new_obs_dict, new_info_dict

    def _respawn_single_vehicle(self, randomize_position=False):
        """
        Arbitrary insert a new vehicle to a new spawn place if possible.
        """
        safe_places_dict = self.engine.spawn_manager.get_available_respawn_places(
            self.current_map, randomize=randomize_position
        )
        if len(safe_places_dict) == 0:
            return None, None, None
        born_place_index = get_np_random(self._DEBUG_RANDOM_SEED).choice(list(safe_places_dict.keys()), 1)[0]
        new_spawn_place = safe_places_dict[born_place_index]

        new_agent_id, vehicle, step_info = self.agent_manager.propose_new_vehicle()
        new_spawn_place_config = new_spawn_place["config"]
        new_spawn_place_config = self.engine.spawn_manager.update_destination_for(new_agent_id, new_spawn_place_config)
        vehicle.config.update(new_spawn_place_config)
        vehicle.reset()
        after_step_info = vehicle.after_step()
        step_info.update(after_step_info)
        self.dones[new_agent_id] = False  # Put it in the internal dead-tracking dict.

        new_obs = self.observations[new_agent_id].observe(vehicle)
        return new_agent_id, new_obs, step_info

    def setup_engine(self):
        super(MultiAgentMetaDrive, self).setup_engine()
        self.engine.register_manager("spawn_manager", SpawnManager())


def _test():
    setup_logger(True)
    env = MultiAgentMetaDrive(
        {
            "num_agents": 12,
            "allow_respawn": False,
            "use_render": True,
            "debug": False,
            "manual_control": True,
        }
    )
    o, _ = env.reset()
    total_r = 0
    for i in range(1, 100000):
        # o, r, tm, tc, info = env.step(env.action_space.sample())
        o, r, tm, tc, info = env.step({v_id: [0, 1] for v_id in env.vehicles.keys()})
        for r_ in r.values():
            total_r += r_
        # o, r, tm, tc, info = env.step([0,1])
        # TODO: why does this make sense? total_r is not a vehicle id.
        # d.update({"total_r": total_r})
        # env.render(text=d)
        env.render(mode="top_down")
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    setup_logger(True)
    env = MultiAgentMetaDrive(
        {
            "use_render": True,
            "num_agents": 5,
            "start_seed": 8000,
            "num_scenarios": 1,
            "map": "SSS",
            "vehicle_config": {
                "vehicle_model": "s"
            },
            "target_vehicle_configs": {
                "agent0": {
                    "vehicle_model": "static_default"
                },
                "agent1": {
                    "vehicle_model": "l"
                },
                "agent2": {
                    "vehicle_model": "xl"
                },
            },
            # "allow_respawn": False,
            # "manual_control": True,
        }
    )
    o, _ = env.reset()
    total_r = 0
    for i in range(1, 100000):
        # o, r, tm, tc, info = env.step(env.action_space.sample())
        o, r, tm, tc, info = env.step({v_id: [0.0, 0.0] for v_id in env.vehicles.keys()})
        for r_ in r.values():
            total_r += r_
        # o, r, tm, tc, info = env.step([0,1])
        # tm.update({"total_r": total_r})
        env.render(mode="top_down")
        # env.reset()
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def pygame_replay(
    name,
    env_class,
    save=False,
    other_traj=None,
    film_size=(1000, 1000),
    extra_config={},
):
    import copy
    import json
    import pygame

    extra_config["use_render"] = True
    env = env_class(extra_config)
    ckpt = "metasvodist_{}_best.json".format(name) if other_traj is None else other_traj
    with open(ckpt, "r") as f:
        traj = json.load(f)
    o, _ = env.reset(copy.deepcopy(traj))
    env.main_camera.set_follow_lane(True)
    frame_count = 0
    while True:
        o, r, tm, tc, i = env.step(env.action_space.sample())
        env.engine.force_fps.toggle()
        env.render(mode="top_down", num_stack=50, film_size=film_size, history_smooth=0)
        if save:
            pygame.image.save(
                env._top_down_renderer._runtime_canvas,
                "{}_{}.png".format(name, frame_count),
            )
        frame_count += 1
        if len(env.engine.replay_manager.restore_episode_info) == 0:
            env.close()


def panda_replay(name, env_class, save=False, other_traj=None, extra_config={}):
    import copy
    import json
    import pygame

    extra_config.update({"use_render": True})
    env = env_class(extra_config)
    ckpt = "metasvodist_{}_best.json".format(name) if other_traj is None else other_traj
    with open(ckpt, "r") as f:
        traj = json.load(f)
    o, _ = env.reset(copy.deepcopy(traj))
    env.main_camera.set_follow_lane(True)
    frame_count = 0
    while True:
        o, r, tm, tc, i = env.step(env.action_space.sample())
        env.engine.force_fps.toggle()
        if save:
            pygame.image.save(
                env._top_down_renderer._runtime_canvas,
                "{}_{}.png".format(name, frame_count),
            )
        frame_count += 1
        if len(env.engine.replay_manager.restore_episode_info) == 0:
            env.close()


if __name__ == "__main__":
    # _test()
    _vis()



import pytest
import numpy as np

from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import WaymoIDMPolicy
from metadrive.policy.replay_policy import WaymoReplayEgoCarPolicy


@pytest.mark.parametrize("policy", [WaymoIDMPolicy, WaymoReplayEgoCarPolicy])
def test_waymo_env(policy, render=False, num_scenarios=3):
    WaymoIDMPolicy.NORMAL_SPEED = 30
    asset_path = AssetLoader.asset_path
    try:
        env = WaymoEnv(
            {
                "manual_control": False,
                "no_traffic": True if policy == WaymoIDMPolicy else False,
                "use_render": render,
                "agent_policy": policy,
                "data_directory": AssetLoader.file_path(asset_path, "waymo", return_raw_style=False),
                "num_scenarios": num_scenarios
            }
        )
        for seed in range(0, num_scenarios):
            env.reset(seed=seed)
            for i in range(1000):
                o, r, tm, tc, info = env.step([1.0, 0.])
                if tm or tc:
                    assert info["arrive_dest"], "Can not arrive dest"
                    print("{} track_length: ".format(env.engine.global_seed), info["track_length"])
                    # assert info["arrive_dest"], "Can not arrive dest"
                    break

                if i == 999:
                    raise ValueError("Can not arrive dest")
    finally:
        env.close()


def test_store_map_memory_leakage(render=False):
    WaymoIDMPolicy.NORMAL_SPEED = 30
    asset_path = AssetLoader.asset_path
    env = WaymoEnv(
        {
            "manual_control": False,
            "no_traffic": False,
            "store_map": True,
            "use_render": render,
            "agent_policy": WaymoReplayEgoCarPolicy,
            "data_directory": AssetLoader.file_path(asset_path, "waymo", return_raw_style=False),
            "num_scenarios": 3
        }
    )
    try:

        memory = []
        for _ in range(10):
            # test twp times for testing loading stored map
            for seed in range(3):
                env.reset(seed=seed)
                for i in range(1000):
                    o, r, tm, tc, info = env.step([1.0, 0.])
                    if tm or tc:
                        assert info["arrive_dest"], "Can not arrive dest"
                        assert env.episode_step > 60
                        break
                    if i == 999:
                        raise ValueError("Can not arrive dest")

            def process_memory():
                import psutil
                import os
                process = psutil.Process(os.getpid())
                mem_info = process.memory_info()
                return mem_info.rss

            memory.append(process_memory() / 1024 / 1024)  # in mb
        # print(memory)
        assert np.std(memory) < 2, "They should be almost the same"
    finally:
        env.close()


if __name__ == "__main__":
    test_store_map_memory_leakage(render=True)
    # test_waymo_env(policy=WaymoIDMPolicy, render=True)



from metadrive.envs.marl_envs.marl_bottleneck import MultiAgentBottleneckEnv
from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv
from metadrive.envs.marl_envs.marl_intersection import MultiAgentIntersectionEnv
from metadrive.envs.marl_envs.marl_parking_lot import MultiAgentParkingLotEnv
from metadrive.envs.marl_envs.marl_tollgate import MultiAgentTollgateEnv
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.envs.marl_envs.tinyinter import MultiAgentTinyInter



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

from metadrive.render_pipeline.rplibs.six import iteritems, string_types
from metadrive.render_pipeline.rpcore.rpobject import RPObject

__all__ = ["make_setting_from_data"]


def make_setting_from_factory(data, factory):
    """ Constructs a new setting from a given dataset, alongst with a factory
    to resolve the setting type to """
    if data["type"] not in factory:
        raise Exception("Unkown setting type: {}".format(data["type"]))
    try:
        instance = factory[data["type"]](data)
    except Exception:
        print("Exception occured while parsing", data)
        raise
    if data:
        raise Exception("Unparsed data left in plugin setting: {}".format(data))
    return instance


def make_setting_from_data(data):
    """ Constructs a new setting from a given dataset. This method will automatically
    instantiate a new class matching the type of the given dataset. It will fill
    all values provided by the dataset and then return the created instance """
    factory = {
        "int": IntType,
        "float": FloatType,
        "bool": BoolType,
        "enum": EnumType,
        "path": PathType,
        "power_of_two": PowerOfTwoType,
        "sample_sequence": SampleSequenceType
    }
    return make_setting_from_factory(data, factory)


class BaseType(RPObject):
    """ This is the base setting type, all setting types derive from this """
    def __init__(self, data):
        """ Fills in all settings from the given setting data """
        self.value = None
        self.type = data.pop("type")
        self.label = data.pop("label").strip()
        self.description = data.pop("description").strip()
        self.runtime = data.pop("runtime", False)
        self.shader_runtime = data.pop("shader_runtime", False)
        self.display_conditions = data.pop("display_if", {})

        RPObject.__init__(self, "psetting:{}".format(self.label))

    def set_value(self, value):
        """ Sets the value of the setting, should get overridden """
        raise NotImplementedError()

    def add_defines(self, plugin_id, setting_id, defines):
        """ Makes the value of this plugin available as a define """
        defines["{}_{}".format(plugin_id, setting_id)] = self.value

    def should_be_visible(self, settings):
        """ Evaluates whether the plugin should be visible, taking all display
        conditions into account """
        for key, val in iteritems(self.display_conditions):
            if settings[key].value != val:
                return False
        return True


class TemplatedType(BaseType):
    """ This setting stores a single type including a minimum and maximum value.
    It is shared between integer and floating point types. """
    def __init__(self, template_type, data):
        BaseType.__init__(self, data)
        self.template_type = template_type
        self.default = template_type(data.pop("default"))
        self.value = self.default

        setting_range = data.pop("range")
        self.minval = template_type(setting_range[0])
        self.maxval = template_type(setting_range[1])

    def set_value(self, value):
        value = self.template_type(value)
        if self.minval <= value <= self.maxval:
            self.value = value
        else:
            self.error("Invalid value: {}".format(value))


class IntType(TemplatedType):
    """ Template instantiation of TemplatedType using int """
    def __init__(self, data):
        TemplatedType.__init__(self, int, data)


class PowerOfTwoType(IntType):
    """ Type for any power of two resolution """
    def __init__(self, data):
        IntType.__init__(self, data)

    def set_value(self, value):
        value = int(value)
        if self.minval <= value <= self.maxval:
            if value in [2**i for i in range(32)]:
                self.value = value
            else:
                self.error("Not a power of two: {}".format(value))
        else:
            self.error("Invalid value: {}".format(value))


class FloatType(TemplatedType):
    """ Template instantiation of TemplatedType using float """
    def __init__(self, data):
        TemplatedType.__init__(self, float, data)


class BoolType(BaseType):
    """ Boolean setting type """
    def __init__(self, data):
        BaseType.__init__(self, data)
        self.default = bool(data.pop("default"))
        self.value = self.default

    def set_value(self, value):
        if isinstance(value, string_types):
            self.value = str(value.lower()) in ("true", "1")
        else:
            self.value = bool(value)


class EnumType(BaseType):
    """ Enumeration setting type """
    def __init__(self, data):
        BaseType.__init__(self, data)
        self.values = tuple(data.pop("values"))
        self.default = data.pop("default")
        if self.default not in self.values:
            raise Exception("Enum default not in enum values: {}".format(self.default))
        self.value = self.default

    def set_value(self, value):
        if value not in self.values:
            self.error("Value '" + str(value) + "' not in enum values!")
            return
        self.value = value

    def add_defines(self, plugin_id, setting_id, defines):
        defines["{}_{}".format(plugin_id, setting_id)] = 1000 + self.values.index(self.value)

        for i, val in enumerate(self.values):
            defines["enum_{}_{}_{}".format(plugin_id, setting_id, val)] = 1000 + i


class SampleSequenceType(BaseType):
    """ Type for any 2D or 3D sample sequence """

    POISSON_2D_SIZES = (4, 8, 12, 16, 32, 64)
    POISSON_3D_SIZES = (16, 32, 64)
    HALTON_SIZES = (4, 8, 16, 32, 64, 128)

    def __init__(self, data):
        BaseType.__init__(self, data)
        self.dimension = int(data.pop("dimension"))
        if self.dimension not in (2, 3):
            raise Exception("Not a valid dimension, must be 2 or 3!")
        self.default = data.pop("default")
        if self.default not in self.sequences:
            raise Exception("Not a valid sequence: {}".format(self.default))
        self.value = self.default

    def set_value(self, value):
        if value not in self.sequences:
            self.error("Value '" + str(value) + "' is not a valid sequence!")
            return
        self.value = value

    @property
    def sequences(self):
        result = []
        if self.dimension == 2:
            for dim in self.POISSON_2D_SIZES:
                result.append("poisson_2D_" + str(dim))
        else:
            for dim in self.POISSON_3D_SIZES:
                result.append("poisson_3D_" + str(dim))
        for dim in self.HALTON_SIZES:
            result.append("halton_" + str(self.dimension) + "D_" + str(dim))
        return result


class PathType(BaseType):
    """ Path type to specify paths to files """
    def __init__(self, data):
        BaseType.__init__(self, data)
        self.default = str(data.pop("default"))
        self.value = self.default
        self.file_type = str(data.pop("file_type"))
        self.base_path = str(data.pop("base_path"))

    def set_value(self, value):
        self.value = str(value)

    def add_defines(self, *args):
        # Paths are not available in shaders
        pass



# import numpy

import math

import numpy as np
from panda3d.bullet import BulletRigidBodyNode, BulletPlaneShape
from panda3d.bullet import ZUp, BulletHeightfieldShape
from panda3d.core import SamplerState, PNMImage, CardMaker, LQuaternionf
from panda3d.core import Vec3, ShaderTerrainMesh, Texture, TextureStage

from metadrive.base_class.base_object import BaseObject
from metadrive.constants import CamMask
from metadrive.constants import MetaDriveType, CollisionGroup
from metadrive.engine.asset_loader import AssetLoader
from metadrive.third_party.diamond_square import diamond_square


class Terrain(BaseObject):
    COLLISION_MASK = CollisionGroup.Terrain
    HEIGHT = 0.0
    PROBE_HEIGHT = 600
    PROBE_SIZE = 1024

    def __init__(self, show_terrain, engine):
        use_render_pipeline = engine.use_render_pipeline
        super(Terrain, self).__init__(random_seed=0)

        # collision mesh
        self.simple_terrain_collision_mesh = None  # a flat collision shape
        self.terrain_collision_mesh = None  # a 3d mesh

        # visualization mesh feature
        heightfield_image_size = 4096  # fixed image size 4096*4096
        self._height_scale = engine.global_config["height_scale"]  # [m]
        self._drivable_region_extension = engine.global_config["drivable_region_extension"]  # [m] road marin
        self._terrain_size = 2048  # [m]
        self._semantic_map_size = 512  # [m] it should include the whole map. Otherwise, road will have no texture!
        self._semantic_map_pixel_per_meter = 22  # [m] how many pixels per meter
        # pre calculate some variables
        self._downsample_rate = int(heightfield_image_size / self._terrain_size)  # downsample to 2048 m
        self._elevation_texture_ratio = self._terrain_size / self._semantic_map_size  # for shader

        self._mesh_terrain = None
        self._mesh_terrain_height = None
        self._mesh_terrain_node = None
        self._terrain_shader_set = False  # only set once
        self.probe = None

        if self.render and show_terrain:
            if engine.use_render_pipeline:
                self._load_mesh_terrain_textures()
                self._mesh_terrain_node = ShaderTerrainMesh()
                # disable env probe as some vehicle models may break it
                # self.probe = engine.render_pipeline.add_environment_probe()
                # self.probe.set_pos(0, 0, self.PROBE_HEIGHT)
                # self.probe.set_scale(self.PROBE_SIZE * 2, self.PROBE_SIZE * 2, 1000)
            else:
                self._generate_card_terrain()

    def _load_mesh_terrain_textures(self, anisotropic_degree=16, minfilter=SamplerState.FT_linear_mipmap_linear):
        """
        Only maintain one copy of all asset
        :param anisotropic_degree: https://docs.panda3d.org/1.10/python/programming/texturing/texture-filter-types
        :param minfilter: https://docs.panda3d.org/1.10/python/programming/texturing/texture-filter-types
        :return: None
        """
        # texture stage
        self.ts_color = TextureStage("color")
        self.ts_normal = TextureStage("normal")
        self.ts_normal.setMode(TextureStage.M_normal)

        # basic height_field
        heightfield_tex = self.loader.loadTexture(AssetLoader.file_path("textures", "terrain", "heightfield.png"))
        heightfield_img = np.frombuffer(heightfield_tex.getRamImage().getData(), dtype=np.uint16)
        self.heightfield_img = heightfield_img.reshape((heightfield_tex.getYSize(), heightfield_tex.getXSize(), 1))

        # grass
        self.grass_tex = self.loader.loadTexture(
            AssetLoader.file_path("textures", "grass2", "grass_path_2_diff_1k.png")
        )
        self.grass_normal = self.loader.loadTexture(
            AssetLoader.file_path("textures", "grass2", "grass_path_2_nor_gl_1k.png")
        )
        self.grass_rough = self.loader.loadTexture(
            AssetLoader.file_path("textures", "grass2", "grass_path_2_rough_1k.png")
        )

        v_wrap = Texture.WMRepeat
        u_warp = Texture.WMMirror
        filter_type = Texture.FTLinearMipmapLinear
        anisotropic_degree = 16
        for tex in [self.grass_tex, self.grass_normal, self.grass_rough]:
            tex.set_wrap_u(u_warp)
            tex.set_wrap_v(v_wrap)
            tex.setMinfilter(filter_type)
            tex.setMagfilter(filter_type)
            tex.setAnisotropicDegree(anisotropic_degree)

        # rock
        self.rock_tex = self.loader.loadTexture(
            AssetLoader.file_path("textures", "rock", "brown_mud_leaves_01_diff_1k.png")
        )
        self.rock_normal = self.loader.loadTexture(
            AssetLoader.file_path("textures", "rock", "brown_mud_leaves_01_nor_gl_1k.png")
        )
        self.rock_rough = self.loader.loadTexture(
            AssetLoader.file_path("textures", "rock", "brown_mud_leaves_01_rough_1k.png")
        )

        v_wrap = Texture.WMRepeat
        u_warp = Texture.WMMirror
        filter_type = Texture.FTLinearMipmapLinear
        anisotropic_degree = 16
        for tex in [self.rock_tex, self.rock_normal, self.rock_rough]:
            tex.set_wrap_u(u_warp)
            tex.set_wrap_v(v_wrap)
            tex.setMinfilter(filter_type)
            tex.setMagfilter(filter_type)
            tex.setAnisotropicDegree(anisotropic_degree)

        # Road surface
        # self.road_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "sci", "new_color.png"))
        self.road_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "asphalt", "diff_2k.png"))
        self.road_texture_normal = self.loader.loadTexture(
            AssetLoader.file_path("textures", "asphalt", "normal_2k.png")
        )
        self.road_texture_rough = self.loader.loadTexture(AssetLoader.file_path("textures", "asphalt", "rough_2k.png"))
        v_wrap = Texture.WMRepeat
        u_warp = Texture.WMMirror
        filter_type = Texture.FTLinearMipmapLinear
        anisotropic_degree = 16
        for tex in [self.road_texture_rough, self.road_texture, self.road_texture_normal]:
            tex.set_wrap_u(u_warp)
            tex.set_wrap_v(v_wrap)
            tex.setMinfilter(filter_type)
            tex.setMagfilter(filter_type)
            tex.setAnisotropicDegree(anisotropic_degree)
        # self.road_texture_displacement = self.loader.loadTexture(AssetLoader.file_path("textures", "sci", "normal.jpg"))
        # self.road_texture.setMinfilter(minfilter)
        # self.road_texture.setAnisotropicDegree(anisotropic_degree)

        # lane line
        white_lane_line = PNMImage(1024, 1024, 4)
        white_lane_line.fill(1., 1., 1.)
        self.white_lane_line = Texture("white lane line")
        self.white_lane_line.load(white_lane_line)

        yellow_lane_line = PNMImage(1024, 1024, 4)
        yellow_lane_line.fill(*(255 / 255, 200 / 255, 0 / 255))
        self.yellow_lane_line = Texture("white lane line")
        self.yellow_lane_line.load(yellow_lane_line)

    # @time_me
    def _generate_mesh_vis_terrain(
        self,
        size,
        heightfield: Texture,
        attribute_tex: Texture,
        target_triangle_width=10,
        height_scale=100,
        height_offset=0.,
        engine=None,
    ):
        """
        Given a height field map to generate terrain and an attribute_tex to texture terrain, so we can get road/grass
        pixels_per_meter is determined by heightfield.size/size
        lane line and so on.
        :param size: [m] this terrain of the generate terrain
        :param heightfield: terrain heightfield. It should be a 16-bit png and have a quadratic size of a power of two.
        :param attribute_tex: doing texture splatting. r,g,b,a represent: grass/road/lane_line ratio respectively
        :param target_triangle_width: For a value of 10.0 for example, the terrain will attempt to make every triangle 10 pixels wide on screen.
        :param height_scale: Scale the height of mountain.
        :param engine: engine instance
        :return:
        """
        engine = engine or self.engine
        assert engine is not None
        assert isinstance(heightfield, Texture), "heightfield file must be Panda3D Texture"

        # Set a heightfield,
        heightfield.wrap_u = SamplerState.WM_clamp
        heightfield.wrap_v = SamplerState.WM_clamp
        self._mesh_terrain_node.heightfield = heightfield

        # Set the target triangle width.
        self._mesh_terrain_node.target_triangle_width = target_triangle_width

        # Generate the terrain
        self._mesh_terrain_node.generate()
        self._mesh_terrain = self.origin.attach_new_node(self._mesh_terrain_node)
        self._set_terrain_shader(engine, attribute_tex)

        # Attach the terrain to the main scene and set its scale. With no scale
        # set, the terrain ranges from (0, 0, 0) to (1, 1, 1)
        self._mesh_terrain.set_scale(size, size, height_scale)
        self._mesh_terrain_height = height_offset
        self._mesh_terrain.set_pos(-size / 2, -size / 2, self._mesh_terrain_height)

    def _set_terrain_shader(self, engine, attribute_tex):
        """
        Set a shader on the terrain. The ShaderTerrainMesh only works with an applied shader. You can use the shaders
        used here in your own application.

        Note: you have to make sure you modified the terrain_effect.yaml and vert.glsl/frag.glsl together, as they are
        made for different render pipeline. We expect the same terrain generated from different pipelines.
        """
        if engine.use_render_pipeline and not self._terrain_shader_set:
            engine.render_pipeline.reload_shaders()
            terrain_effect = AssetLoader.file_path("effect", "terrain_effect.yaml")
            engine.render_pipeline.set_effect(self._mesh_terrain, terrain_effect, {}, 100)
            # # height
            self._mesh_terrain.set_shader_input("height_scale", self._height_scale)

            # grass
            self._mesh_terrain.set_shader_input("grass_tex", self.grass_tex)
            self._mesh_terrain.set_shader_input("grass_normal", self.grass_normal)
            self._mesh_terrain.set_shader_input("grass_rough", self.grass_rough)

            # road
            self._mesh_terrain.set_shader_input("rock_tex", self.rock_tex)
            self._mesh_terrain.set_shader_input("rock_normal", self.rock_normal)
            self._mesh_terrain.set_shader_input("rock_rough", self.rock_rough)

            self._mesh_terrain.set_shader_input("road_tex", self.road_texture)
            self._mesh_terrain.set_shader_input("yellow_tex", self.yellow_lane_line)
            self._mesh_terrain.set_shader_input("white_tex", self.white_lane_line)
            self._mesh_terrain.set_shader_input("road_normal", self.road_texture_normal)
            self._mesh_terrain.set_shader_input("road_rough", self.road_texture_rough)
            self._mesh_terrain.set_shader_input("elevation_texture_ratio", self._elevation_texture_ratio)
            self._terrain_shader_set = True
        self._mesh_terrain.set_shader_input("attribute_tex", attribute_tex)

    def reset(self, center_position):
        """
        Update terrain according to current map
        """
        assert self.engine is not None, "Can not call this without initializing engine"

        # if not self.use_render_pipeline and self.simple_terrain_collision_mesh is None:
        # TODO: I disabled online terrain collision mesh generation now, consider enabling it in the future
        if self.simple_terrain_collision_mesh is None:
            self.detach_from_world(self.engine.physics_world)
            # If no render pipeline, we can only have 2d terrain. It will only be generated for once.
            shape = BulletPlaneShape(Vec3(0, 0, 1), -0.05)
            node = BulletRigidBodyNode(MetaDriveType.GROUND)
            node.setFriction(.9)
            node.addShape(shape)

            node.setIntoCollideMask(self.COLLISION_MASK)
            self.dynamic_nodes.append(node)

            self.simple_terrain_collision_mesh = self.origin.attachNewNode(node)
            self._node_path_list.append(np)
            self.attach_to_world(self.engine.render, self.engine.physics_world)

        # elif self.use_render_pipeline:
        if self.use_render_pipeline:
            self.detach_from_world(self.engine.physics_world)
            assert self.engine.current_map is not None, "Can not find current map"
            semantics = self.engine.current_map.get_semantic_map(
                size=self._semantic_map_size,
                pixels_per_meter=self._semantic_map_pixel_per_meter,
                polyline_thickness=int(1024 / self._semantic_map_size),
                layer=["lane", "lane_line"]
            )
            semantics = semantics.astype(np.float32)
            semantic_tex = Texture()
            semantic_tex.setup2dTexture(*semantics.shape[:2], Texture.TFloat, Texture.FRgba)
            semantic_tex.setRamImage(semantics)

            # we will downsmaple the precision after this
            drivable_region = self.engine.current_map.get_height_map(
                self._terrain_size, self._downsample_rate, self._drivable_region_extension
            )
            drivable_region_height = np.mean(self.heightfield_img[np.where(drivable_region)]).astype(np.uint16)
            heightfield_img = np.where(drivable_region, drivable_region_height, self.heightfield_img)

            # set to zero height
            heightfield_img -= drivable_region_height

            # down sample
            heightfield_img = np.array(heightfield_img[::self._downsample_rate, ::self._downsample_rate])
            heightfield = heightfield_img

            heightfield_tex = Texture()
            heightfield_tex.setup2dTexture(*heightfield.shape[:2], Texture.TShort, Texture.FLuminance)
            heightfield_tex.setRamImage(heightfield)

            # # update collision every time!
            # TODO: I disabled online terrain collision mesh generation now, consider enabling it in the future
            # self._generate_collision_mesh(heightfield_img, self.height_scale)
            self._generate_mesh_vis_terrain(
                self._terrain_size, heightfield_tex, semantic_tex, height_scale=self._height_scale, height_offset=0
            )
            self.attach_to_world(self.engine.render, self.engine.physics_world)

        self.set_position(center_position)

    def _generate_collision_mesh(self, heightfield_img, height_scale):
        # TODO we can do some optimization here, only update some regions
        # clear previous mesh
        self.dynamic_nodes.clear()
        mesh = np.zeros([heightfield_img.shape[0] + 1, heightfield_img.shape[1] + 1, 1])
        mesh[:heightfield_img.shape[0], :heightfield_img.shape[1]] = heightfield_img
        mesh = mesh.astype(np.uint16)

        heightfield_tex = Texture()
        heightfield_tex.setup2dTexture(*mesh.shape[:2], Texture.TShort, Texture.FLuminance)
        heightfield_tex.setRamImage(mesh)

        p = PNMImage()
        heightfield_tex.store(p)

        shape = BulletHeightfieldShape(p, height_scale, ZUp)
        shape.setUseDiamondSubdivision(True)

        node = BulletRigidBodyNode(MetaDriveType.GROUND)
        node.setFriction(.9)
        node.addShape(shape)

        node.setIntoCollideMask(CollisionGroup.Terrain)
        self.dynamic_nodes.append(node)

        self.terrain_collision_mesh = self.origin.attachNewNode(node)
        self._node_path_list.append(np)

    def set_position(self, position, height=None):
        if self.render:
            if self.use_render_pipeline:
                pos = (self._mesh_terrain.get_pos()[0] + position[0], self._mesh_terrain.get_pos()[1] + position[1])
                self._mesh_terrain.set_pos(*pos, self._mesh_terrain_height)
                if self.probe is not None:
                    self.probe.set_pos(*pos, self.PROBE_HEIGHT)
            else:
                super(Terrain, self).set_position(position, height)

    def _generate_card_terrain(self):
        self.origin.hide(CamMask.MiniMap | CamMask.Shadow | CamMask.DepthCam | CamMask.ScreenshotCam)
        # self.terrain_normal = self.loader.loadTexture(
        #     AssetLoader.file_path( "textures", "grass2", "normal.jpg")
        # )
        self.terrain_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "ground.png"))
        self.terrain_texture.set_format(Texture.F_srgb)
        self.terrain_texture.setWrapU(Texture.WM_repeat)
        self.terrain_texture.setWrapV(Texture.WM_repeat)
        self.ts_color = TextureStage("color")
        self.ts_normal = TextureStage("normal")
        self.ts_normal.set_mode(TextureStage.M_normal)
        # self.set_position((0, 0), self.HEIGHT)
        cm = CardMaker('card')
        scale = 4000
        cm.setUvRange((0, 0), (scale / 10, scale / 10))
        card = self.origin.attachNewNode(cm.generate())

        self._node_path_list.append(card)

        card.set_scale(scale)
        card.setPos(-scale / 2, -scale / 2, -0.1)
        card.setZ(-.05)
        card.setTexture(self.ts_color, self.terrain_texture)
        # card.setTexture(self.ts_normal, self.terrain_normal)
        self.terrain_texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
        self.terrain_texture.setAnisotropicDegree(8)
        card.setQuat(LQuaternionf(math.cos(-math.pi / 4), math.sin(-math.pi / 4), 0, 0))

    def _make_random_terrain(self, texture_size, terrain_size, heightfield):
        height_1 = width_2 = height_3 = width_3 = length = int((terrain_size - texture_size) / 2)
        width_1 = height_2 = width = texture_size
        max_height = 8192
        min_height = 0
        roughness = 0.14

        array_1 = diamond_square(
            [height_1, width_1], min_height, max_height, roughness, random_seed=self.generate_seed()
        )
        array_2 = diamond_square(
            [height_2, width_2], min_height, max_height, roughness, random_seed=self.generate_seed()
        )

        array_3 = diamond_square(
            [height_3, width_3], min_height, max_height, roughness, random_seed=self.generate_seed()
        )

        heightfield[:length, length:length + width] = array_1
        heightfield[-length:, length:length + width] = array_1

        heightfield[length:length + width, :length] = array_2
        heightfield[length:length + width, -length:] = array_2

        heightfield[:length, :length] = array_3
        heightfield[-length:, :length] = array_3
        heightfield[:length, -length:] = array_3
        heightfield[-length:, -length:] = array_3


# Some useful threads
# GeoMipTerrain:
# https://discourse.panda3d.org/t/texture-mapping-according-with-vertex-position-terrain-generation/28929
# https://github.com/someoneetc/Panda3dTerrainGenerator
# https://docs.panda3d.org/1.10/python/programming/texturing/texture-combine-modes
# texture splatting: https://discourse.panda3d.org/t/developing-a-city-builder-terrain-engine/7331
# https://discourse.panda3d.org/t/advice-about-geomipterrains/8346 optimize geomipterrain efficiency
#
# repeat texture
# # root = terrain.getRoot()
# # root.setTexScale(TextureStage.getDefault(), num_repeats_x, num_repeats_y)

# https://discourse.panda3d.org/t/geomipterrain-and-pnmimages/13728
# use shaderterrain with bullet: https://discourse.panda3d.org/t/getting-shaderterrainmesh-and-bulletheightfieldshape-to-match/27512
# shaderterrain+multi_texture: https://discourse.panda3d.org/t/alpha-texture-in-shaderterrainmesh-not-work/23715/3
# shaderTerrain height texture https://discourse.panda3d.org/t/how-get-height-of-shader-terrain-mesh-for-add-grass-tree-in-map/23964

# multi-thread pssm
# https://discourse.panda3d.org/t/parallel-split-shadow-mapping-using-pssmcamerarig/27457/12



import pickle

from tqdm import tqdm

from metadrive.component.map.pg_map import PGMap, MapGenerateMethod
from metadrive.manager.base_manager import BaseManager
from metadrive.utils.utils import get_time_str


class PGMapManager(BaseManager):
    """
    MapManager contains a list of PGmaps
    """
    PRIORITY = 0  # Map update has the most high priority

    def __init__(self):
        super(PGMapManager, self).__init__()
        self.current_map = None

        # for pgmaps
        start_seed = self.start_seed = self.engine.global_config["start_seed"]
        env_num = self.env_num = self.engine.global_config["num_scenarios"]
        self.maps = {_seed: None for _seed in range(start_seed, start_seed + env_num)}

    def spawn_object(self, object_class, *args, **kwargs):
        # Note: Map instance should not be reused / recycled.
        map = self.engine.spawn_object(object_class, auto_fill_random_seed=False, force_spawn=True, *args, **kwargs)
        self.engine._spawned_objects.pop(map.id)
        return map

    def load_map(self, map):
        map.attach_to_world()
        self.current_map = map

    def unload_map(self, map):
        map.detach_from_world()
        self.current_map = None
        if not self.engine.global_config["store_map"]:
            map.destroy()

    def destroy(self):
        self.maps = None
        super(PGMapManager, self).destroy()

    def before_reset(self):
        # remove map from world before adding
        if self.current_map is not None:
            map = self.current_map
            self.unload_map(map)

    def reset(self):
        config = self.engine.global_config.copy()
        current_seed = self.engine.global_seed

        if self.maps[current_seed] is None:
            map_config = config["map_config"]
            map_config.update({"seed": current_seed})
            map_config = self.add_random_to_map(map_config)
            map = self.spawn_object(PGMap, map_config=map_config, random_seed=None)
            if self.engine.global_config["store_map"]:
                self.maps[current_seed] = map
        else:
            map = self.maps[current_seed]
        self.load_map(map)

    def add_random_to_map(self, map_config):
        if self.engine.global_config["random_lane_width"]:
            map_config[PGMap.LANE_WIDTH
                       ] = self.np_random.rand() * (PGMap.MAX_LANE_WIDTH - PGMap.MIN_LANE_WIDTH) + PGMap.MIN_LANE_WIDTH
        if self.engine.global_config["random_lane_num"]:
            map_config[PGMap.LANE_NUM] = self.np_random.randint(PGMap.MIN_LANE_NUM, PGMap.MAX_LANE_NUM + 1)
        return map_config

    def generate_all_maps(self):
        """
        Call this function to generate all maps before using them
        """
        for seed in tqdm(self.maps.keys(), desc="Generate maps"):
            config = self.engine.global_config.copy()
            current_seed = seed
            self.engine.seed(seed)
            if self.maps[current_seed] is None:
                map_config = config["map_config"]
                map_config.update({"seed": current_seed})
                map_config = self.add_random_to_map(map_config)
                map = self.spawn_object(PGMap, map_config=map_config, random_seed=None)
                self.maps[current_seed] = map
                map.detach_from_world()

    def dump_all_maps(self, file_name=None):
        """
        Dump all maps. If some maps are not generated, we will generate it at first
        """
        if file_name is None:
            start_seed = self.engine.global_config["start_seed"]
            end_seed = start_seed + self.engine.global_config["num_scenarios"]
            file_name = "{}_{}_{}.json".format(start_seed, end_seed, get_time_str())
        self.generate_all_maps()
        ret = {}
        for seed, map in tqdm(self.maps.items(), desc="Dump maps"):
            ret[seed] = map.get_meta_data()
        with open(file_name, "wb+") as file:
            pickle.dump(ret, file)
        return ret

    def load_all_maps(self, file_name):
        if self.current_map is not None:
            self.unload_map(self.current_map)
        with open(file_name, "rb") as file:
            loaded_map_data = pickle.load(file)
        map_seeds = list(loaded_map_data.keys())
        start_seed = min(map_seeds)
        map_num = len(map_seeds)
        assert self.env_num == map_num and start_seed == self.engine.global_config[
            "start_seed"
        ], "The environment num and start seed in config: {}, {} must be the same as the env num and start seed: {}, {} in the loaded file".format(
            self.env_num, self.start_seed, map_num, start_seed
        )

        for i in tqdm(range(self.env_num), desc="Load maps"):
            loaded_seed = i + start_seed
            map_data = loaded_map_data[loaded_seed]
            block_sequence = map_data["block_sequence"]
            map_config = map_data["map_config"]
            map_config[PGMap.GENERATE_TYPE] = MapGenerateMethod.PG_MAP_FILE
            map_config[PGMap.GENERATE_CONFIG] = block_sequence
            map = self.spawn_object(PGMap, map_config=map_config, random_seed=None)
            self.maps[i + self.start_seed] = map
            map.detach_from_world()
        self.reset()
        return loaded_map_data

    def clear_objects(self, *args, **kwargs):
        """
        As Map instance should not be recycled, we will forcefully destroy useless map instances.
        """
        return super(PGMapManager, self).clear_objects(force_destroy=True, *args, **kwargs)


# For compatibility
MapManager = PGMapManager



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# pylint: disable=invalid-name

# This file includes all modules from the native module.

from __future__ import print_function
import sys
from os.path import dirname, realpath

from direct.stdpy.file import join, isfile
from metadrive.render_pipeline.rpcore.rpobject import RPObject

# Store a global flag, indicating whether the C++ modules were loaded or the python
# implemetation of them
NATIVE_CXX_LOADED = True

# Read the configuration from the flag-file, LQY: Bypass this check process.
# current_path = dirname(realpath(__file__))
# cxx_flag_path = join(current_path, "use_cxx.flag")
# if not isfile(cxx_flag_path):
#     RPObject.global_error("CORE", "Could not find cxx flag, please run the setup.py!")
#     sys.exit(1)
# else:
#     with open(join(current_path, "use_cxx.flag"), "r") as handle:
#         NATIVE_CXX_LOADED = handle.read().strip() == "1"

# The native module should only be imported once, and that by the internal pipeline code
assert __package__ == "metadrive.render_pipeline.rpcore.native", "You have included the pipeline in the wrong way!"

# Classes which should get imported
classes_to_import = [
    "GPUCommand",
    "GPUCommandList",
    "ShadowManager",
    "InternalLightManager",
    "PSSMCameraRig",
    "IESDataset",
    "TagStateManager",
]

# Classes which should get imported and renamed
classes_to_import_and_rename = {"RPPointLight": "PointLight", "RPSpotLight": "SpotLight"}

native_module = None

# If the module was built, use it, otherwise use the python wrappers
if NATIVE_CXX_LOADED:
    try:
        from panda3d import _rplight as _native_module  # pylint: disable=wrong-import-position
        # RPObject.global_debug("CORE", "Using panda3d-supplied core module")
    except ImportError:
        # RPObject.global_debug("CORE", "Using native core module")
        from metadrive.render_pipeline.rpcore.native import native_ as _native_module  # pylint: disable=wrong-import-position
else:
    from metadrive.render_pipeline.rpcore import pynative as _native_module  # pylint: disable=wrong-import-position
    RPObject.global_debug("CORE", "Using simulated python-wrapper module")

# Import all classes
for v in classes_to_import + list(classes_to_import_and_rename.keys()):
    if hasattr(_native_module, v):
        v_name = classes_to_import_and_rename[v] if v in classes_to_import_and_rename else v
        globals()[v_name] = getattr(_native_module, v)
    else:
        print("ERROR: could not import class", v, "from", _native_module.__name__)

# Don't export all variables, only the required ones
__all__ = classes_to_import + list(classes_to_import_and_rename.values()) + ["NATIVE_CXX_LOADED"]



from typing import List, Tuple, Union

import numpy as np

from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.lane.straight_lane import StraightLane
from metadrive.constants import PGLineType, PGLineColor
from metadrive.utils.utils import import_pygame
from metadrive.type import MetaDriveType

PositionType = Union[Tuple[float, float], np.ndarray]
pygame = import_pygame()
COLOR_BLACK = pygame.Color("black")


class ObservationWindow:
    def __init__(self, max_range, resolution):
        self.max_range = max_range
        self.resolution = resolution
        self.receptive_field = None
        self.receptive_field_double = None

        self.canvas_rotate = None
        self.canvas_uncropped = pygame.Surface(
            (int(self.resolution[0] * np.sqrt(2)) + 1, int(self.resolution[1] * np.sqrt(2)) + 1)
        )

        self.canvas_display = pygame.Surface(self.resolution)
        self.canvas_display.fill(COLOR_BLACK)

    def reset(self, canvas_runtime):
        canvas_runtime.fill(COLOR_BLACK)

        # Assume max_range is only the radius!
        self.receptive_field_double = (
            int(canvas_runtime.pix(self.max_range[0] * np.sqrt(2))) * 2,
            int(canvas_runtime.pix(self.max_range[1] * np.sqrt(2))) * 2
        )
        self.receptive_field = (
            int(canvas_runtime.pix(self.max_range[0])) * 2, int(canvas_runtime.pix(self.max_range[1])) * 2
        )
        self.canvas_rotate = pygame.Surface(self.receptive_field_double)
        self.canvas_rotate.fill(COLOR_BLACK)
        self.canvas_display.fill(COLOR_BLACK)
        self.canvas_uncropped.fill(COLOR_BLACK)

    def _blit(self, canvas, position):
        self.canvas_rotate.blit(
            canvas, (0, 0), (
                position[0] - self.receptive_field_double[0] / 2, position[1] - self.receptive_field_double[1] / 2,
                self.receptive_field_double[0], self.receptive_field_double[1]
            )
        )

    def _rotate(self, heading):
        rotation = np.rad2deg(heading) + 90
        scale = self.canvas_uncropped.get_size()[0] / self.canvas_rotate.get_size()[0]
        return pygame.transform.rotozoom(self.canvas_rotate, rotation, scale)

    def _crop(self, new_canvas):
        size = self.canvas_display.get_size()
        self.canvas_display.blit(
            new_canvas,
            (0, 0),
            (
                new_canvas.get_size()[0] / 2 - size[0] / 2,  # Left
                new_canvas.get_size()[1] / 2 - size[1] / 2,  # Top
                size[0],  # Width
                size[1]  # Height
            )
        )

    def render(self, canvas, position, heading):
        # Prepare a runtime canvas for rotation. Assume max_range is only the radius, not diameter!
        self._blit(canvas, position)

        # Rotate the image so that ego is always heading top
        new_canvas = self._rotate(heading)

        # Crop the rotated image and then resize to the desired resolution
        self._crop(new_canvas)

        return self.canvas_display

    def get_observation_window(self):
        return self.canvas_display

    def get_size(self):
        assert self.canvas_rotate is not None
        return self.canvas_rotate.get_size()

    def get_screen_window(self):
        return self.get_observation_window()


class WorldSurface(pygame.Surface):
    """
    A pygame Surface implementing a local coordinate system so that we can move and zoom in the displayed area.
    From highway-env, See more information on its Github page: https://github.com/eleurent/highway-env.
    """

    BLACK = (0, 0, 0)
    GREY = (100, 100, 100)
    GREEN = (50, 200, 0)
    YELLOW = (200, 200, 0)
    WHITE = (255, 255, 255)
    INITIAL_SCALING = 5.5
    INITIAL_CENTERING = [0.5, 0.5]
    SCALING_FACTOR = 1.3
    MOVING_FACTOR = 0.1
    LANE_LINE_COLOR = (35, 35, 35)

    def __init__(self, size: Tuple[int, int], flags: object, surf: pygame.SurfaceType) -> None:
        surf.fill(pygame.Color("Black"))
        super().__init__(size, flags, surf)
        self.raw_size = size
        self.raw_flags = flags
        self.raw_surface = surf
        self.origin = np.array([0, 0])
        self.scaling = self.INITIAL_SCALING
        self.centering_position = self.INITIAL_CENTERING
        self.fill(self.BLACK)

    def pix(self, length: float) -> int:
        """
        Convert a distance [m] to pixels [px].

        :param length: the input distance [m]
        :return: the corresponding size [px]
        """
        return int(length * self.scaling)

    def pos2pix(self, x: float, y: float) -> Tuple[int, int]:
        """
        Convert two world coordinates [m] into a position in the surface [px]

        :param x: x world coordinate [m]
        :param y: y world coordinate [m]
        :return: the coordinates of the corresponding pixel [px]
        """
        return self.pix(x - self.origin[0]), self.pix(y - self.origin[1])

    def vec2pix(self, vec: PositionType) -> Tuple[int, int]:
        """
        Convert a world position [m] into a position in the surface [px].

        :param vec: a world position [m]
        :return: the coordinates of the corresponding pixel [px]
        """
        return self.pos2pix(vec[0], vec[1])

    def is_visible(self, vec: PositionType, margin: int = 50) -> bool:
        """
        Is a position visible in the surface?
        :param vec: a position
        :param margin: margins around the frame to test for visibility
        :return: whether the position is visible
        """
        x, y = self.vec2pix(vec)
        return -margin < x < self.get_width() + margin and -margin < y < self.get_height() + margin

    def move_display_window_to(self, position: PositionType) -> None:
        """
        Set the origin of the displayed area to center on a given world position.

        :param position: a world position [m]
        """
        self.origin = position - np.array(
            [
                self.centering_position[0] * self.get_width() / self.scaling,
                self.centering_position[1] * self.get_height() / self.scaling
            ]
        )

    def copy(self):
        ret = WorldSurface(size=self.raw_size, flags=self.raw_flags, surf=self.raw_surface)
        ret.origin = self.origin
        ret.scaling = self.scaling
        ret.centering_position = self.centering_position
        ret.blit(self, (0, 0))
        return ret


class VehicleGraphics:
    RED = (255, 100, 100)
    GREEN = (50, 200, 0)
    BLUE = (100, 200, 255)
    YELLOW = (200, 200, 0)
    BLACK = (60, 60, 60)
    PURPLE = (200, 0, 150)
    DEFAULT_COLOR = YELLOW
    EGO_COLOR = GREEN
    font = None

    @classmethod
    def display(
        cls, vehicle, surface, color, heading, label: bool = False, draw_countour=False, contour_width=1
    ) -> None:
        """
        Display a vehicle on a pygame surface.

        The vehicle is represented as a colored rotated rectangle.

        :param vehicle: the vehicle to be drawn
        :param surface: the surface to draw the vehicle on
        :param label: whether a text label should be rendered
        """
        if not surface.is_visible(vehicle.position):
            return
        w = surface.pix(vehicle.WIDTH)
        h = surface.pix(vehicle.LENGTH)
        position = [*surface.pos2pix(vehicle.position[0], vehicle.position[1])]
        angle = np.rad2deg(heading)
        box = [pygame.math.Vector2(p) for p in [(-h / 2, -w / 2), (-h / 2, w / 2), (h / 2, w / 2), (h / 2, -w / 2)]]
        box_rotate = [p.rotate(angle) + position for p in box]

        pygame.draw.polygon(surface, color, box_rotate)
        if draw_countour and pygame.ver.startswith("2"):
            pygame.draw.polygon(surface, cls.BLACK, box_rotate, width=contour_width)  # , 1)

        # Label
        if label:
            if cls.font is None:
                cls.font = pygame.font.Font(None, 15)
            text = "#{}".format(id(vehicle) % 1000)
            text = cls.font.render(text, 1, (10, 10, 10), (255, 255, 255))
            surface.blit(text, position)

    @classmethod
    def get_color(cls, vehicle) -> Tuple[int]:
        if vehicle.crashed:
            color = cls.RED
        else:
            color = cls.BLUE
        return color


class LaneGraphics:
    """A visualization of a lane."""

    STRIPE_SPACING: float = 5
    """ Offset between stripes [m]"""

    STRIPE_LENGTH: float = 3
    """ Length of a stripe [m]"""

    STRIPE_WIDTH: float = 0.3
    """ Width of a stripe [m]"""

    LANE_LINE_WIDTH: float = 1

    @classmethod
    def display(cls, lane, surface, two_side=True, use_line_color=False, color=None) -> None:
        """
        Display a lane on a surface.

        :param lane: the lane to be displayed
        :param surface: the pygame surface
        :param two_side: draw two sides of the lane, or only one side
        """
        side = 2 if two_side else 1
        stripes_count = int(2 * (surface.get_height() + surface.get_width()) / (cls.STRIPE_SPACING * surface.scaling))
        s_origin, _ = lane.local_coordinates(surface.origin)
        s0 = (int(s_origin) // cls.STRIPE_SPACING - stripes_count // 2) * cls.STRIPE_SPACING
        for side in range(side):
            if use_line_color:
                if lane.line_colors[side] == PGLineColor.YELLOW and lane.line_types[side] == PGLineType.CONTINUOUS:
                    color = (0, 80, 220)
                elif lane.line_types[side] == PGLineType.SIDE:
                    color = (160, 160, 160)
                else:
                    color = (80, 80, 80)
            if lane.line_types[side] == PGLineType.BROKEN:
                cls.striped_line(lane, surface, stripes_count, s0, side, color=color)
            # circular side or continuous, it is same now
            elif lane.line_types[side] == PGLineType.CONTINUOUS and isinstance(lane, CircularLane):
                cls.continuous_curve(lane, surface, stripes_count, s0, side, color=color)
            elif lane.line_types[side] == PGLineType.SIDE and isinstance(lane, CircularLane):
                cls.continuous_curve(lane, surface, stripes_count, s0, side, color=color)
            # the line of continuous straight and side straight is same now
            elif (lane.line_types[side] == PGLineType.CONTINUOUS) and isinstance(lane, StraightLane):
                cls.continuous_line(lane, surface, stripes_count, s0, side, color=color)
            elif (lane.line_types[side] == PGLineType.SIDE) and isinstance(lane, StraightLane):
                cls.continuous_line(lane, surface, stripes_count, s0, side, color=color)
            # special scenario
            elif lane.line_types[side] == PGLineType.NONE:
                continue
            else:
                raise ValueError("I don't know how to draw this line type: {}".format(lane.line_types[side]))

    @classmethod
    def display_scenario(cls, waymo_poly_line, type, surface) -> None:
        """
        Display a lane on a surface.

        :param lane: the lane to be displayed
        :param surface: the pygame surface
        :param two_side: draw two sides of the lane, or only one side
        """
        lane = waymo_poly_line
        if MetaDriveType.is_yellow_line(type):
            color = (0, 80, 220)
        elif MetaDriveType.is_road_edge(type):
            color = (160, 160, 160)
        else:
            color = (80, 80, 80)
        if MetaDriveType.is_road_line(type) or MetaDriveType.is_road_edge(type):
            # if len(waymo_poly_line.segment_property) < 1:
            #     return
            stripes_count = int(
                2 * (surface.get_height() + surface.get_width()) / (cls.STRIPE_SPACING * surface.scaling)
            )
            s_origin, _ = lane.local_coordinates(surface.origin)
            s0 = (int(s_origin) // cls.STRIPE_SPACING - stripes_count // 2) * cls.STRIPE_SPACING

            if MetaDriveType.is_broken_line(type):
                starts = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING
                ends = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_LENGTH
                lats = [0 for s in starts]
            else:
                starts = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING
                ends = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_SPACING
                lats = [0 for s in starts]
            cls.draw_stripes(lane, surface, starts, ends, lats, color=color)
        elif type == "center_lane" or type is None:
            pass

    @classmethod
    def display_nuplan(cls, poly_line, type, color, surface) -> None:
        """
        Display a lane on a surface.
        """
        lane = poly_line
        if color == PGLineColor.YELLOW:
            color = (0, 80, 220)
        else:
            color = (80, 80, 80)
        if len(poly_line.segment_property) <= 1:
            return
        stripes_count = int(2 * (surface.get_height() + surface.get_width()) / (cls.STRIPE_SPACING * surface.scaling))
        s_origin, _ = lane.local_coordinates(surface.origin)
        s0 = (int(s_origin) // cls.STRIPE_SPACING - stripes_count // 2) * cls.STRIPE_SPACING

        if type == PGLineType.BROKEN:
            starts = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING
            ends = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_LENGTH
            lats = [0 for s in starts]
        else:
            starts = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING
            ends = s0 + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_SPACING
            lats = [0 for s in starts]
        cls.draw_stripes(lane, surface, starts, ends, lats, color=color)

    @classmethod
    def striped_line(cls, lane, surface, stripes_count: int, longitudinal: float, side: int, color=None) -> None:
        """
        Draw a striped line on one side of a lane, on a surface.

        :param lane: the lane
        :param surface: the pygame surface
        :param stripes_count: the number of stripes to draw
        :param longitudinal: the longitudinal position of the first stripe [m]
        :param side: which side of the road to draw [0:left, 1:right]
        """
        starts = longitudinal + np.arange(stripes_count) * cls.STRIPE_SPACING
        ends = longitudinal + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_LENGTH
        lats = [(side - 0.5) * lane.width_at(s) for s in starts]
        cls.draw_stripes(lane, surface, starts, ends, lats, color=color)

    @classmethod
    def continuous_curve(cls, lane, surface, stripes_count: int, longitudinal: float, side: int, color=None) -> None:
        """
        Draw a striped line on one side of a lane, on a surface.

        :param lane: the lane
        :param surface: the pygame surface
        :param stripes_count: the number of stripes to draw
        :param longitudinal: the longitudinal position of the first stripe [m]
        :param side: which side of the road to draw [0:left, 1:right]
        """
        starts = longitudinal + np.arange(stripes_count) * cls.STRIPE_SPACING
        ends = longitudinal + np.arange(stripes_count) * cls.STRIPE_SPACING + cls.STRIPE_SPACING
        lats = [(side - 0.5) * lane.width_at(s) for s in starts]
        cls.draw_stripes(lane, surface, starts, ends, lats, color=color)

    @classmethod
    def continuous_line(cls, lane, surface, stripes_count: int, longitudinal: float, side: int, color=None) -> None:
        """
        Draw a continuous line on one side of a lane, on a surface.

        :param lane: the lane
        :param surface: the pygame surface
        :param stripes_count: the number of stripes that would be drawn if the line was striped
        :param longitudinal: the longitudinal position of the start of the line [m]
        :param side: which side of the road to draw [0:left, 1:right]
        """
        starts = [longitudinal + 0 * cls.STRIPE_SPACING]
        ends = [longitudinal + stripes_count * cls.STRIPE_SPACING + cls.STRIPE_LENGTH]
        lats = [(side - 0.5) * lane.width_at(s) for s in starts]
        cls.draw_stripes(lane, surface, starts, ends, lats, color=color)

    @classmethod
    def draw_stripes(cls, lane, surface, starts: List[float], ends: List[float], lats: List[float], color=None) -> None:
        """
        Draw a set of stripes along a lane.

        :param lane: the lane
        :param surface: the surface to draw on
        :param starts: a list of starting longitudinal positions for each stripe [m]
        :param ends: a list of ending longitudinal positions for each stripe [m]
        :param lats: a list of lateral positions for each stripe [m]
        """
        if color is None:
            color = surface.LANE_LINE_COLOR
        starts = np.clip(starts, 0, lane.length)
        ends = np.clip(ends, 0, lane.length)
        for k, _ in enumerate(starts):
            if abs(starts[k] - ends[k]) > 0.5 * cls.STRIPE_LENGTH:
                pygame.draw.line(
                    surface, color, (surface.vec2pix(lane.position(starts[k], lats[k]))),
                    (surface.vec2pix(lane.position(ends[k], lats[k]))),
                    max(surface.pix(cls.STRIPE_WIDTH), surface.pix(cls.LANE_LINE_WIDTH))
                )

    @classmethod
    def draw_drivable_area(cls, lane, surface, color=(255, 255, 255)):
        from metadrive.component.pgblock.pg_block import PGBlock
        segment_num = int(lane.length / PGBlock.LANE_SEGMENT_LENGTH)
        width = lane.width
        for segment in range(segment_num):
            p_1 = lane.position(segment * PGBlock.LANE_SEGMENT_LENGTH, -width / 2)
            p_2 = lane.position(segment * PGBlock.LANE_SEGMENT_LENGTH, width / 2)
            p_3 = lane.position((segment + 1) * PGBlock.LANE_SEGMENT_LENGTH, width / 2)
            p_4 = lane.position((segment + 1) * PGBlock.LANE_SEGMENT_LENGTH, -width / 2)
            pygame.draw.polygon(
                surface, color,
                [surface.pos2pix(*p_1),
                 surface.pos2pix(*p_2),
                 surface.pos2pix(*p_3),
                 surface.pos2pix(*p_4)]
            )

        # # for last part
        p_1 = lane.position(segment_num * PGBlock.LANE_SEGMENT_LENGTH, -width / 2)
        p_2 = lane.position(segment_num * PGBlock.LANE_SEGMENT_LENGTH, width / 2)
        p_3 = lane.position(lane.length, width / 2)
        p_4 = lane.position(lane.length, -width / 2)
        pygame.draw.polygon(
            surface, color,
            [surface.pos2pix(*p_1),
             surface.pos2pix(*p_2),
             surface.pos2pix(*p_3),
             surface.pos2pix(*p_4)]
        )


class ObservationWindowMultiChannel:
    CHANNEL_NAMES = ["road_network", "traffic_flow", "target_vehicle", "past_pos"]

    def __init__(self, names, max_range, resolution):
        assert isinstance(names, list)
        assert set(self.CHANNEL_NAMES)
        self.sub_observations = {
            k: ObservationWindow(max_range=max_range, resolution=resolution)
            for k in ["traffic_flow", "target_vehicle"]
        }
        self.sub_observations["road_network"] = ObservationWindow(
            max_range=max_range,
            resolution=(resolution[0] * 2, resolution[1] * 2)
            # max_range=max_range, resolution=resolution
        )

        self.resolution = (resolution[0] * 2, resolution[1] * 2)
        self.canvas_display = None

    def get_canvas_display(self):
        if self.canvas_display is None:
            self.canvas_display = pygame.Surface(self.resolution)
        self.canvas_display.fill(COLOR_BLACK)
        return self.canvas_display

    def reset(self, canvas_runtime):
        for k, sub in self.sub_observations.items():
            sub.reset(canvas_runtime)

    def render(self, canvas_dict, position, heading):
        assert isinstance(canvas_dict, dict)
        assert set(canvas_dict.keys()) == set(self.sub_observations.keys())
        ret = dict()
        for k, canvas in canvas_dict.items():
            ret[k] = self.sub_observations[k].render(canvas, position, heading)
        return self.get_observation_window(ret)

    def get_observation_window(self, canvas_dict=None):
        if canvas_dict is None:
            canvas_dict = {k: v.get_observation_window() for k, v in self.sub_observations.items()}
        return canvas_dict

    def get_size(self):
        return next(iter(self.sub_observations.values())).get_size()

    def get_screen_window(self):
        canvas = self.get_canvas_display()
        ret = self.get_observation_window()

        for k in ret.keys():
            if k == "road_network":
                continue
            ret[k] = pygame.transform.scale2x(ret[k])

        def _draw(canvas, key, color):
            mask = pygame.mask.from_threshold(ret[key], (0, 0, 0, 0), (10, 10, 10, 255))
            mask.to_surface(canvas, setcolor=None, unsetcolor=color)

        if "navigation" in ret:
            _draw(canvas, "navigation", pygame.Color("Blue"))
        _draw(canvas, "road_network", pygame.Color("White"))
        _draw(canvas, "traffic_flow", pygame.Color("Red"))
        _draw(canvas, "target_vehicle", pygame.Color("Green"))
        return canvas



import gymnasium as gym

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

setup_logger(debug=True)


def _test_destroy(config):
    env = MetaDriveEnv(config)
    try:
        env.reset()
        for i in range(1, 20):
            env.step([1, 1])

        env.close()
        env.reset()
        env.close()
        env.close()
        env.reset()
        env.reset()
        env.close()

        # Again!
        env = MetaDriveEnv(config)
        env.reset()
        for i in range(1, 20):
            env.step([1, 1])
        env.reset()
        env.close()
    finally:
        env.close()


def test_discrete_action():
    # Close and reset
    env = MetaDriveEnv(
        dict(
            discrete_action=True,
            use_multi_discrete=False,
            discrete_steering_dim=3,
            discrete_throttle_dim=5,
            action_check=True
        )
    )
    try:
        env.reset()
        assert isinstance(env.action_space, gym.spaces.Discrete)
        assert env.action_space.n == 15
        v = env.vehicle
        policy = env.engine.get_policy(v.name)
        assert policy.convert_to_continuous_action(0) == (-1, -1)
        assert policy.convert_to_continuous_action(1) == (0, -1)
        assert policy.convert_to_continuous_action(2) == (1, -1)
        assert policy.convert_to_continuous_action(7) == (0, 0)
        assert policy.convert_to_continuous_action(14) == (1, 1)

        for _ in range(20):
            o, r, tm, tc, i = env.step(env.action_space.sample())

    finally:
        env.close()


def test_multi_discrete_action():
    # Close and reset
    env = MetaDriveEnv(
        dict(
            discrete_action=True,
            use_multi_discrete=True,
            discrete_steering_dim=3,
            discrete_throttle_dim=5,
            action_check=True
        )
    )
    try:
        env.reset()
        assert isinstance(env.action_space, gym.spaces.MultiDiscrete)
        assert env.action_space.shape == (2, )
        assert all(env.action_space.nvec == (3, 5))
        v = env.vehicle
        policy = env.engine.get_policy(v.name)
        assert policy.convert_to_continuous_action([0, 0]) == (-1, -1)
        assert policy.convert_to_continuous_action([1, 0]) == (0, -1)
        assert policy.convert_to_continuous_action([2, 0]) == (1, -1)
        assert policy.convert_to_continuous_action([1, 2]) == (0, 0)
        assert policy.convert_to_continuous_action([2, 4]) == (1, 1)

        for _ in range(20):
            o, r, tm, tc, i = env.step(env.action_space.sample())

    finally:
        env.close()


if __name__ == "__main__":
    test_discrete_action()
    test_multi_discrete_action()



"""
A unified data format to describe a scenario that can be replayed by MetaDrive Simulator.

Example:

    scenario = {

        # ===== Meta data about the scenario =====
        # string. The name of the scenario
        "id": "Waymo-001",

        # string. The version of data format.
        "version": "MetaDrive v0.3.0.1",


        # int. The length of all trajectory and state arrays (T).
        "length": 200,

        # ===== Meta data ===
        "metadata": {

            # np.ndarray in (T, ). The time stamp of each time step.
            "ts": np.array([0.0, 0.1, 0.2, ...], dtype=np.float32),


            # bool. Whether the scenario is processed and exported by MetaDrive.
            # Some operations may be done, such as interpolating the lane to
            # make way points uniformly scattered in given interval.
            "metadrive_processed": True,

            # string. Coordinate system.
            "coordinate": "metadrive",

            # optional keys
            "source_file": "training_20s.tfrecord-00014-of-01000",
            "dataset": "waymo",
            "scenario_id": "dd0c8c27fdd6ef59",  # Used in Waymo dataset
            "seed": 512,
            "history_metadata": {},

            "sdc_id": "172",  # A key exists in tracks

        },

        # ===== Trajectories of active participants, e.g. vehicles, pedestrians =====
        # a dict mapping object ID to it's state dict.
        "tracks": {
            "vehicle1": {

                # The type string in metadrive.scenario.MetaDriveType
                "type": "VEHICLE",

                # The state dict. All values must have T elements.
                "state": {
                    "position": np.ones([200, 3], dtype=np.float32),
                    ...
                },

                # The meta data dict. Store useful information about the object. type in metadata could be those from
                # different dataset
                "metadata": {
                    "type": "VEHICLE",
                    "track_length": 200,
                    "object_id": "vehicle1",

                    # Optional keys
                    "agent_name": "default_agent",
                    "policy_spawn_info": {  # Information needed to re-instantiate the policy
                        "policy_class": ("metadrive.policy.idm_policy", "IDMPolicy"),
                        "args": ...,
                        "kwargs": ...,
                    }
                }
            },

            "pedestrian1": ...
        },

        # ===== States sequence of dynamics objects, e.g. traffic light =====
        # a dict mapping object ID to it's state dict.
        "dynamic_map_states": {
            "trafficlight1": {

                # The type string in metadrive.scenario.MetaDriveType
                "type": "TRAFFIC_LIGHT",

                # The state dict. All values must have T elements.
                "state": {
                    "object_state": np.ones([200, ], dtype=int),
                    ...
                },

                # The meta data dict. Store useful information about the object
                "metadata": {
                    "type": "TRAFFIC_LIGHT",
                    "track_length": 200,
                }
        }

        # ===== Map features =====
        # A dict mapping from map feature ID to a line segment
        "map_features": {
            "219": {
                "type": "LANE_SURFACE_STREET",
                "polyline": np.array in [21, 2],  # A set of 2D points describing a line segment
                # optional, only works for lane
                "polygon": np.array in [N, 2] # A set of 2D points representing convexhull
            },
            "182": ...
            ...
        }
    }
"""
import math
import os
from collections import defaultdict

import numpy as np

from metadrive.type import MetaDriveType


class ScenarioDescription(dict):
    """
    MetaDrive Scenario Description. It stores keys of the data dict.
    """
    TRACKS = "tracks"
    VERSION = "version"
    ID = "id"
    DYNAMIC_MAP_STATES = "dynamic_map_states"
    MAP_FEATURES = "map_features"
    LENGTH = "length"
    METADATA = "metadata"
    FIRST_LEVEL_KEYS = {TRACKS, VERSION, ID, DYNAMIC_MAP_STATES, MAP_FEATURES, LENGTH, METADATA}

    # lane keys
    POLYLINE = "polyline"
    POLYGON = "polygon"
    LEFT_BOUNDARIES = "left_boundaries"
    RIGHT_BOUNDARIES = "right_boundaries"
    LEFT_NEIGHBORS = "left_neighbor"
    RIGHT_NEIGHBORS = "right_neighbor"
    ENTRY = "entry_lanes"
    EXIT = "exit_lanes"

    TYPE = "type"
    STATE = "state"
    OBJECT_ID = "object_id"
    STATE_DICT_KEYS = {TYPE, STATE, METADATA}
    ORIGINAL_ID_TO_OBJ_ID = "original_id_to_obj_id"
    OBJ_ID_TO_ORIGINAL_ID = "obj_id_to_original_id"
    TRAFFIC_LIGHT_POSITION = "stop_point"
    TRAFFIC_LIGHT_STATUS = "object_state"
    TRAFFIC_LIGHT_LANE = "lane"

    METADRIVE_PROCESSED = "metadrive_processed"
    TIMESTEP = "ts"
    COORDINATE = "coordinate"
    SDC_ID = "sdc_id"  # Not necessary, but can be stored in metadata.
    METADATA_KEYS = {METADRIVE_PROCESSED, COORDINATE, TIMESTEP}

    ALLOW_TYPES = (int, float, str, np.ndarray, dict, list, tuple, type(None), set)

    class SUMMARY:
        OBJECT_SUMMARY = "object_summary"
        NUMBER_SUMMARY = "number_summary"

        # for each object summary
        TYPE = "type"
        OBJECT_ID = "object_id"
        TRACK_LENGTH = "track_length"
        MOVING_DIST = "moving_distance"
        VALID_LENGTH = "valid_length"
        CONTINUOUS_VALID_LENGTH = "continuous_valid_length"

        # for number summary:
        OBJECT_TYPES = "object_types"
        NUM_OBJECTS = "num_objects"
        NUM_MOVING_OBJECTS = "num_moving_objects"
        NUM_OBJECTS_EACH_TYPE = "num_objects_each_type"
        NUM_MOVING_OBJECTS_EACH_TYPE = "num_moving_objects_each_type"

        NUM_TRAFFIC_LIGHTS = "num_traffic_lights"
        NUM_TRAFFIC_LIGHT_TYPES = "num_traffic_light_types"
        NUM_TRAFFIC_LIGHTS_EACH_STEP = "num_traffic_light_each_step"

        NUM_MAP_FEATURES = "num_map_features"
        MAP_HEIGHT_DIFF = "map_height_diff"

    class DATASET:
        SUMMARY_FILE = "dataset_summary.pkl"  # dataset summary file name
        MAPPING_FILE = "dataset_mapping.pkl"  # store the relative path of summary file and each scenario

    @classmethod
    def sanity_check(cls, scenario_dict, check_self_type=False, valid_check=False):

        if check_self_type:
            assert isinstance(scenario_dict, dict)
            assert not isinstance(scenario_dict, ScenarioDescription)

        # Whether input has all required keys
        assert cls.FIRST_LEVEL_KEYS.issubset(set(scenario_dict.keys())), \
            "You lack these keys in first level: {}".format(cls.FIRST_LEVEL_KEYS.difference(set(scenario_dict.keys())))

        # Check types, only native python objects
        # This is to avoid issue in pickle deserialization
        _recursive_check_type(scenario_dict, cls.ALLOW_TYPES)

        scenario_length = scenario_dict[cls.LENGTH]

        # Check tracks data
        assert isinstance(scenario_dict[cls.TRACKS], dict)
        for obj_id, obj_state in scenario_dict[cls.TRACKS].items():
            cls._check_object_state_dict(
                obj_state, scenario_length=scenario_length, object_id=obj_id, valid_check=valid_check
            )

        # Check dynamic_map_state
        assert isinstance(scenario_dict[cls.DYNAMIC_MAP_STATES], dict)
        for obj_id, obj_state in scenario_dict[cls.DYNAMIC_MAP_STATES].items():
            cls._check_object_state_dict(obj_state, scenario_length=scenario_length, object_id=obj_id)

        # Check metadata
        assert isinstance(scenario_dict[cls.METADATA], dict)
        assert cls.METADATA_KEYS.issubset(set(scenario_dict[cls.METADATA].keys())), \
            "You lack these keys in metadata: {}".format(
                cls.METADATA_KEYS.difference(set(scenario_dict[cls.METADATA].keys()))
            )
        assert scenario_dict[cls.METADATA][cls.TIMESTEP].shape == (scenario_length, )

    @classmethod
    def _check_object_state_dict(cls, obj_state, scenario_length, object_id, valid_check=True):
        # Check keys
        assert set(obj_state).issuperset(cls.STATE_DICT_KEYS)

        # Check type
        assert MetaDriveType.has_type(obj_state[cls.TYPE]
                                      ), "MetaDrive doesn't have this type: {}".format(obj_state[cls.TYPE])

        # Check set type
        assert obj_state[cls.TYPE] != MetaDriveType.UNSET, "Types should be set for objects and traffic lights"

        # Check state arrays temporal consistency
        assert isinstance(obj_state[cls.STATE], dict)
        for state_key, state_array in obj_state[cls.STATE].items():
            assert isinstance(state_array, (np.ndarray, list, tuple))
            assert len(state_array) == scenario_length

            if not isinstance(state_array, np.ndarray):
                continue

            assert state_array.ndim in [1, 2], "Haven't implemented test array with dim {} yet".format(state_array.ndim)
            if state_array.ndim == 2:
                assert state_array.shape[
                    1] != 0, "Please convert all state with dim 1 to a 1D array instead of 2D array."

            if state_key == "valid" and valid_check:
                assert np.sum(state_array) >= 1, "No frame valid for this object. Consider removing it"

            # check valid
            if "valid" in obj_state[cls.STATE] and valid_check:
                _array = state_array[..., :2] if state_key == "position" else state_array
                assert abs(np.sum(_array[np.where(obj_state[cls.STATE]["valid"], False, True)])) < 1e-2, \
                    "Valid array mismatches with {} array, some frames in {} have non-zero values, " \
                    "so it might be valid".format(state_key, state_key)

        # Check metadata
        assert isinstance(obj_state[cls.METADATA], dict)
        for metadata_key in (cls.TYPE, cls.OBJECT_ID):
            assert metadata_key in obj_state[cls.METADATA]

        # Check metadata alignment
        if cls.OBJECT_ID in obj_state[cls.METADATA]:
            assert obj_state[cls.METADATA][cls.OBJECT_ID] == object_id

    def to_dict(self):
        return dict(self)

    def get_sdc_track(self):
        assert self.SDC_ID in self[self.METADATA]
        sdc_id = str(self[self.METADATA][self.SDC_ID])
        return self[self.TRACKS][sdc_id]

    @staticmethod
    def get_object_summary(state_dict, id):
        type = state_dict["type"]
        state_dict = state_dict["state"]
        track = state_dict["position"]
        valid_track = track[np.where(state_dict["valid"].astype(int))][..., :2]
        distance = float(
            sum(np.linalg.norm(valid_track[i] - valid_track[i + 1]) for i in range(valid_track.shape[0] - 1))
        )
        valid_length = int(sum(state_dict["valid"]))

        continuous_valid_length = 0
        for v in state_dict["valid"]:
            if v:
                continuous_valid_length += 1
            if continuous_valid_length > 0 and not v:
                break

        return {
            ScenarioDescription.SUMMARY.TYPE: type,
            ScenarioDescription.SUMMARY.OBJECT_ID: str(id),
            ScenarioDescription.SUMMARY.TRACK_LENGTH: int(len(track)),
            ScenarioDescription.SUMMARY.MOVING_DIST: float(distance),
            ScenarioDescription.SUMMARY.VALID_LENGTH: int(valid_length),
            ScenarioDescription.SUMMARY.CONTINUOUS_VALID_LENGTH: int(continuous_valid_length)
        }

    @staticmethod
    def get_export_file_name(dataset, dataset_version, scenario_name):
        return "sd_{}_{}_{}.pkl".format(dataset, dataset_version, scenario_name)

    @staticmethod
    def is_scenario_file(file_name):
        file_name = os.path.basename(file_name)
        assert file_name[-4:] == ".pkl", "{} is not .pkl file".format(file_name)
        file_name = file_name.replace(".pkl", "")
        return os.path.basename(file_name)[:3] == "sd_" or all(char.isdigit() for char in file_name)

    @staticmethod
    def _calculate_num_moving_objects(scenario):
        # moving object
        number_summary_dict = {
            ScenarioDescription.SUMMARY.NUM_MOVING_OBJECTS: 0,
            ScenarioDescription.SUMMARY.NUM_MOVING_OBJECTS_EACH_TYPE: defaultdict(int)
        }
        for v in scenario[ScenarioDescription.METADATA][ScenarioDescription.SUMMARY.OBJECT_SUMMARY].values():
            if v[ScenarioDescription.SUMMARY.MOVING_DIST] > 1:
                number_summary_dict[ScenarioDescription.SUMMARY.NUM_MOVING_OBJECTS] += 1
                number_summary_dict[ScenarioDescription.SUMMARY.NUM_MOVING_OBJECTS_EACH_TYPE][v["type"]] += 1
        return number_summary_dict

    @staticmethod
    def get_number_summary(scenario):
        number_summary_dict = {}

        # object
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_OBJECTS] = len(scenario[ScenarioDescription.TRACKS])
        number_summary_dict[ScenarioDescription.SUMMARY.OBJECT_TYPES
                            ] = set(v["type"] for v in scenario[ScenarioDescription.TRACKS].values())
        object_types_counter = defaultdict(int)
        for v in scenario[ScenarioDescription.TRACKS].values():
            object_types_counter[v["type"]] += 1
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_OBJECTS_EACH_TYPE] = dict(object_types_counter)

        # moving object
        number_summary_dict.update(ScenarioDescription._calculate_num_moving_objects(scenario))

        # Number of different dynamic object states
        dynamic_object_states_types = set()
        dynamic_object_states_counter = defaultdict(int)
        for v in scenario[ScenarioDescription.DYNAMIC_MAP_STATES].values():
            for step_state in v["state"]["object_state"]:
                if step_state is None:
                    continue
                dynamic_object_states_types.add(step_state)
                dynamic_object_states_counter[step_state] += 1
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_TRAFFIC_LIGHTS
                            ] = len(scenario[ScenarioDescription.DYNAMIC_MAP_STATES])
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_TRAFFIC_LIGHT_TYPES] = dynamic_object_states_types
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_TRAFFIC_LIGHTS_EACH_STEP
                            ] = dict(dynamic_object_states_counter)

        # map
        number_summary_dict[ScenarioDescription.SUMMARY.NUM_MAP_FEATURES
                            ] = len(scenario[ScenarioDescription.MAP_FEATURES])
        number_summary_dict[ScenarioDescription.SUMMARY.MAP_HEIGHT_DIFF
                            ] = ScenarioDescription.map_height_diff(scenario[ScenarioDescription.MAP_FEATURES])
        return number_summary_dict

    @staticmethod
    def sdc_moving_dist(scenario):
        SD = ScenarioDescription
        metadata = scenario[SD.METADATA]
        sdc_info = metadata[SD.SUMMARY.OBJECT_SUMMARY][metadata[SD.SDC_ID]]
        moving_dist = sdc_info[SD.SUMMARY.MOVING_DIST]
        return moving_dist

    @staticmethod
    def num_object(scenario, object_type=None):
        SD = ScenarioDescription
        metadata = scenario[SD.METADATA]
        if object_type is None:
            return metadata[SD.SUMMARY.NUMBER_SUMMARY][SD.SUMMARY.NUM_OBJECTS]
        else:
            return metadata[SD.SUMMARY.NUMBER_SUMMARY][SD.SUMMARY.NUM_OBJECTS_EACH_TYPE].get(object_type, 0)

    @staticmethod
    def num_moving_object(scenario, object_type=None):
        SD = ScenarioDescription
        metadata = scenario[SD.METADATA]
        if SD.SUMMARY.NUM_MOVING_OBJECTS not in metadata[SD.SUMMARY.NUMBER_SUMMARY]:
            num_summary = SD._calculate_num_moving_objects(scenario)
        else:
            num_summary = metadata[SD.SUMMARY.NUMBER_SUMMARY]

        if object_type is None:
            return num_summary[SD.SUMMARY.NUM_MOVING_OBJECTS]
        else:
            return num_summary[SD.SUMMARY.NUM_MOVING_OBJECTS_EACH_TYPE].get(object_type, 0)

    @staticmethod
    def map_height_diff(map_features, target=10):
        max = -math.inf
        min = math.inf
        for feature in map_features.values():
            if not MetaDriveType.is_road_line(feature[ScenarioDescription.TYPE]):
                continue
            polyline = feature[ScenarioDescription.POLYLINE]
            if len(polyline[0]) == 3:
                z = np.asarray(polyline)[..., -1]
                z_max = np.max(z)
                if z_max > max:
                    max = z_max
                z_min = np.min(z)
                if z_min < min:
                    min = z_min
            if max - min > target:
                break
        return float(max - min)


def _recursive_check_type(obj, allow_types, depth=0):
    if isinstance(obj, dict):
        for k, v in obj.items():
            assert isinstance(k, str), "Must use string to be dict keys"
            _recursive_check_type(v, allow_types, depth=depth + 1)

    if isinstance(obj, list):
        for v in obj:
            _recursive_check_type(v, allow_types, depth=depth + 1)

    assert isinstance(obj, allow_types), "Object type {} not allowed! ({})".format(type(obj), allow_types)

    if depth > 1000:
        raise ValueError()


# TODO (LQY): Remove me after paper writing
# {
#     "map_features": {
#         "map_object_id_1": {
#             "center_line": [...],
#             "polygon": [...],
#             "from_lanes": [...],
#             "to_lanes": [...],
#             "neighbor_lanes": [...],
#             "metadata": {...}
#         },
#         "map_object_id_2": {
#             "polyline": [...],
#             "type": "white_solid",
#             "metadata": {...}
#         }
#     },
#     "objects": {
#         "object_id_1": {
#             "position": [...],
#             "velocity": [...],
#             "heading": [...],
#             "size": [l, w, h],
#             "valid": [...],
#             "type": "VEHICLE",
#             "metadata": {...}
#         }
#     },
#     "traffic light": {
#         "light_id_1": {
#             "states": [...],
#             "position": [x, y, z],
#             "heading": -np.pi,
#             "lane_id": "lane_id_1",
#             "metadata": {...}
#         }
#     },
#     "metadata": {
#         "dataset": "nuscenes",
#         "episode_length": 198,
#         "time_interval": 0.1,
#         "sdc_id": "ego",
#         "coordinates": "right-hand"
#         ...
#     }
# }
#
# {
#     "map_features": {
#         "map_object_id_1": {
#             "center_line": [...],
#             "polygon": [...],
#             "connectivity": {...}
#         },
#         "map_object_id_2": {
#             "polyline": [...],
#             "type": "white_solid",
#         }
#     },
#     "objects": {
#         "object_id_1": {
#             "position": [...],
#             "heading": [...],
#             "type": "VEHICLE",
#         }
#     },
#     "traffic light": {
#         "light_id_1": {
#             "states": [...],
#             "lane_id": "lane_id_1",
#         }
#     },
#     "metadata": {
#         "dataset": "nuscenes",
#         "time_interval": 0.1,
#         "sdc_id": "ego",
#     }
# }
#
# {"map_features": {
# "map_object_id_1": {
#     "center_line": [...],
#     "type": "lane",
#     "connectivity": {...}},
# "map_object_id_2": {
#     "polyline": [...],
#     "type": "white_solid_line"}
# },
# "objects": {
#     "object_id_1": {
#         "position": [...],
#         "heading": [...],
#         "type": "VEHICLE"}
# },
# "traffic light": {
#     "light_id_1": {
#         "states": [...],
#         "lane": "map_object_id_1"}
# },
# "metadata": {
#     "dataset": "nuscenes",
#     "time_interval": 0.1
# }}



import logging

import numpy as np

from metadrive.constants import DrivableAreaProperty
import math

from metadrive.component.lane.point_lane import PointLane
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils.math import norm, mph_to_kmh
from metadrive.scenario.utils import read_scenario_data


# return the nearest point"s index of the line
def nearest_point(point, line):
    dist = np.square(line - point)
    dist = np.sqrt(dist[:, 0] + dist[:, 1])
    return np.argmin(dist)


class ScenarioLane(PointLane):
    VIS_LANE_WIDTH = 6
    MAX_SPEED_LIMIT = 100

    def __init__(self, lane_id: int, map_data: dict, need_lane_localization):
        """
        Extract the lane information of one lane, and do coordinate shift if required
        """
        center_line_points = np.asarray(map_data[lane_id][ScenarioDescription.POLYLINE])
        if ScenarioDescription.POLYGON in map_data[lane_id] and len(map_data[lane_id][ScenarioDescription.POLYGON]) > 3:
            polygon = np.asarray(map_data[lane_id][ScenarioDescription.POLYGON])
        else:
            polygon = None
        if "speed_limit_kmh" in map_data[lane_id] or "speed_limit_mph" in map_data[lane_id]:
            speed_limit_kmh = map_data[lane_id].get("speed_limit_kmh", None)
            if speed_limit_kmh is None:
                speed_limit_kmh = mph_to_kmh(map_data[lane_id]["speed_limit_mph"])
        else:
            speed_limit_kmh = self.MAX_SPEED_LIMIT
        super(ScenarioLane, self).__init__(
            center_line_points=center_line_points,
            width=self.get_lane_width(lane_id, map_data),
            polygon=polygon,
            speed_limit=speed_limit_kmh,
            need_lane_localization=need_lane_localization
        )
        self.index = lane_id
        self.lane_type = map_data[lane_id]["type"]
        self.entry_lanes = map_data[lane_id].get(ScenarioDescription.ENTRY, None)
        self.exit_lanes = map_data[lane_id].get(ScenarioDescription.EXIT, None)
        self.left_lanes = map_data[lane_id].get(ScenarioDescription.LEFT_NEIGHBORS, None)
        self.right_lanes = map_data[lane_id].get(ScenarioDescription.RIGHT_NEIGHBORS, None)

    @staticmethod
    def try_get_polygon(map_data, lane_id):
        """
        This method tries to infer polygon from the boundaries
        """
        raise DeprecationWarning("This function can not infer the polygon still")
        lane_info = map_data[lane_id]
        center_line = lane_info["polyline"]
        left_boundary_points = []
        for boundary_info in lane_info[ScenarioDescription.LEFT_BOUNDARIES]:
            boundary = map_data[boundary_info["boundary_feature_id"]][ScenarioDescription.POLYLINE]
            start = center_line[int(boundary_info["lane_start_index"])]
            end = center_line[int(boundary_info["lane_end_index"])]

            start_on_boundary = nearest_point(start, boundary)
            end_on_boundary = nearest_point(end, boundary)

            points = boundary[start_on_boundary:end_on_boundary]
            left_boundary_points.append(points)

        left_boundary_points = np.concatenate(left_boundary_points, axis=0) if len(left_boundary_points) > 0 else []

        right_boundary_points = []
        for boundary_info in lane_info[ScenarioDescription.RIGHT_BOUNDARIES]:
            boundary = map_data[boundary_info["boundary_feature_id"]][ScenarioDescription.POLYLINE]
            start = center_line[int(boundary_info["lane_start_index"])]
            end = center_line[int(boundary_info["lane_end_index"])]

            start_on_boundary = nearest_point(start, boundary)
            end_on_boundary = nearest_point(end, boundary)

            points = boundary[start_on_boundary:end_on_boundary]
            right_boundary_points.append(points)

        right_boundary_points = np.concatenate(right_boundary_points, axis=0) if len(right_boundary_points) > 0 else []
        if len(left_boundary_points) == 0 or len(right_boundary_points) == 0:
            return None
        else:
            polygon = np.concatenate([left_boundary_points, right_boundary_points], axis=0)[..., :2]
            return np.asarray(polygon)

    def get_lane_width(self, lane_id, map_data):
        """
        We use this function to get possible lane width from raw data
        """
        if not (ScenarioDescription.RIGHT_NEIGHBORS in map_data[lane_id]
                and ScenarioDescription.LEFT_NEIGHBORS in map_data[lane_id]):
            return self.VIS_LANE_WIDTH
        right_lanes = map_data[lane_id][ScenarioDescription.RIGHT_NEIGHBORS]
        left_lanes = map_data[lane_id][ScenarioDescription.LEFT_NEIGHBORS]
        if len(right_lanes) + len(left_lanes) == 0:
            return max(sum(map_data[lane_id]["width"][0]), self.VIS_LANE_WIDTH)
        dist_to_left_lane = 0
        dist_to_right_lane = 0
        if len(right_lanes) > 0 and "feature_id" in right_lanes[0]:
            right_lane = map_data[right_lanes[0]["feature_id"]]
            self_start = int(right_lanes[0]["self_start_index"])
            neighbor_start = int(right_lanes[0]["neighbor_start_index"])
            n_point = right_lane[ScenarioDescription.POLYLINE][neighbor_start]
            self_point = map_data[lane_id][ScenarioDescription.POLYLINE][self_start]
            dist_to_right_lane = norm(n_point[0] - self_point[0], n_point[1] - self_point[1])
        if len(left_lanes) > 0 and "feature_id" in left_lanes[0]:
            left_lane = map_data[left_lanes[-1]["feature_id"]]
            self_start = int(left_lanes[-1]["self_start_index"])
            neighbor_start = int(left_lanes[-1]["neighbor_start_index"])
            n_point = left_lane[ScenarioDescription.POLYLINE][neighbor_start]
            self_point = map_data[lane_id][ScenarioDescription.POLYLINE][self_start]
            dist_to_left_lane = norm(n_point[0] - self_point[0], n_point[1] - self_point[1])
        return max(dist_to_left_lane, dist_to_right_lane, self.VIS_LANE_WIDTH)

    def __del__(self):
        logging.debug("ScenarioLane is released")

    def destroy(self):
        self.index = None
        self.entry_lanes = None
        self.exit_lanes = None
        self.left_lanes = None
        self.right_lanes = None
        super(ScenarioLane, self).destroy()


if __name__ == "__main__":
    file_path = AssetLoader.file_path("waymo", "test.pkl", return_raw_style=False)
    data = read_scenario_data(file_path)
    print(data)
    lane = ScenarioLane(108, data["map_features"])



class PGBlockDistConfig:
    MAX_LANE_NUM = 5
    MIN_LANE_NUM = 1

    # Register the block types here! Set their probability to 0.0 if you don't wish it appears in standard MetaDrive.
    BLOCK_TYPE_DISTRIBUTION_V1 = {
        "Curve": 0.5,
        "Straight": 0.1,
        "StdInterSection": 0.075,
        "Roundabout": 0.05,
        "StdTInterSection": 0.075,
        "InRampOnStraight": 0.1,
        "OutRampOnStraight": 0.1,
        "InFork": 0.00,
        "OutFork": 0.00,
        "Merge": 0.00,
        "Split": 0.00,
        "ParkingLot": 0.00,
        "TollGate": 0.00
    }

    BLOCK_TYPE_DISTRIBUTION_V2 = {
        # 0.3 for curves
        "Curve": 0.3,
        # 0.3 for straight
        "Straight": 0.1,
        "InRampOnStraight": 0.1,
        "OutRampOnStraight": 0.1,
        # 0.3 for intersection
        "StdInterSection": 0.15,
        "StdTInterSection": 0.15,
        # 0.1 for roundabout
        "Roundabout": 0.1,
        "InFork": 0.00,
        "OutFork": 0.00,
        "Merge": 0.00,
        "Split": 0.00,
        "ParkingLot": 0.00,
        "TollGate": 0.00,
        "Bidirection": 0.00
    }

    @classmethod
    def all_blocks(cls, version: str = "v2"):
        ret = list(cls._get_dist(version).keys())
        for k in ret:
            assert isinstance(k, str)
        return ret

    @classmethod
    def get_block(cls, block_id: str, version: str = "v2"):
        from metadrive.utils.registry import get_metadrive_class

        for block in cls.all_blocks(version):
            block = get_metadrive_class(block)
            if block.ID == block_id:
                return block
        raise ValueError("No {} block type".format(block_id))

    @classmethod
    def block_probability(cls, version: str = "v2"):
        return list(cls._get_dist(version).values())

    @classmethod
    def _get_dist(cls, version: str):
        # if version == "v1":
        #     return cls.BLOCK_TYPE_DISTRIBUTION_V1
        if version == "v2":
            return cls.BLOCK_TYPE_DISTRIBUTION_V2
        else:
            raise ValueError("Unknown version: {}".format(version))



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rpcore.rpobject import RPObject

from panda3d.core import GeomVertexFormat, GeomVertexData, Geom, GeomVertexWriter
from panda3d.core import GeomTriangles, GeomNode, NodePath, Camera, OmniBoundingVolume
from panda3d.core import TransparencyAttrib, OrthographicLens, Vec4


class PostProcessRegion(RPObject):
    """ Simple wrapper class to create fullscreen triangles and quads """
    @classmethod
    def make(cls, internal_buffer, *args):
        return cls(internal_buffer, *args)

    def __init__(self, internal_buffer, *args):
        RPObject.__init__(self)
        self._buffer = internal_buffer
        self._region = self._buffer.make_display_region(*args)
        self._node = NodePath("RTRoot")

        self._make_fullscreen_tri()
        self._make_fullscreen_cam()
        self._init_function_pointers()

    def _init_function_pointers(self):
        self.set_sort = self._region.set_sort
        self.set_instance_count = self._tri.set_instance_count
        self.disable_clears = self._region.disable_clears
        self.set_active = self._region.set_active
        self.set_clear_depth_active = self._region.set_clear_depth_active
        self.set_clear_depth = self._region.set_clear_depth
        self.set_shader = self._tri.set_shader
        self.set_camera = self._region.set_camera
        self.set_clear_color_active = self._region.set_clear_color_active
        self.set_clear_color = self._region.set_clear_color
        self.set_attrib = self._tri.set_attrib

    def _make_fullscreen_tri(self):
        """ Creates the oversized triangle used for rendering """
        vformat = GeomVertexFormat.get_v3()
        vdata = GeomVertexData("vertices", vformat, Geom.UH_static)
        vdata.set_num_rows(3)
        vwriter = GeomVertexWriter(vdata, "vertex")
        vwriter.add_data3f(-1, 0, -1)
        vwriter.add_data3f(3, 0, -1)
        vwriter.add_data3f(-1, 0, 3)
        gtris = GeomTriangles(Geom.UH_static)
        gtris.add_next_vertices(3)
        geom = Geom(vdata)
        geom.add_primitive(gtris)
        geom_node = GeomNode("gn")
        geom_node.add_geom(geom)
        geom_node.set_final(True)
        geom_node.set_bounds(OmniBoundingVolume())
        tri = NodePath(geom_node)
        tri.set_depth_test(False)
        tri.set_depth_write(False)
        tri.set_attrib(TransparencyAttrib.make(TransparencyAttrib.M_none), 10000)
        tri.set_color(Vec4(1))
        tri.set_bin("unsorted", 10)
        tri.reparent_to(self._node)
        self._tri = tri

    def set_shader_input(self, *args, **kwargs):
        if kwargs.get("override", False):
            self._node.set_shader_input(*args, priority=100000)
        else:
            self._tri.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        self._tri.set_shader_inputs(**kwargs)

    def _make_fullscreen_cam(self):
        """ Creates an orthographic camera for the buffer """
        buffer_cam = Camera("BufferCamera")
        lens = OrthographicLens()
        lens.set_film_size(2, 2)
        lens.set_film_offset(0, 0)
        lens.set_near_far(-100, 100)
        buffer_cam.set_lens(lens)
        buffer_cam.set_cull_bounds(OmniBoundingVolume())
        self._camera = self._node.attach_new_node(buffer_cam)
        self._region.set_camera(self._camera)



import time

import matplotlib.pyplot as plt
import numpy as np

from metadrive.envs.metadrive_env import MetaDriveEnv
import time

if __name__ == '__main__':
    W, H = 1920, 1200
    config = dict(
        # controller="joystick",

        # use_render=True,
        image_observation=True,
        manual_control=True,  # set false for external subscriber control
        traffic_density=0.0,
        num_scenarios=100,
        random_agent_model=True,
        random_lane_width=True,
        random_lane_num=True,
        vehicle_config=dict(image_source="rgb_camera", rgb_camera=(W, H), stack_size=1),
        map=4,  # seven block
        start_seed=0,
        window_size=(300, 200),
    )
    env = MetaDriveEnv(config)
    start = time.time()
    env.reset()
    frames = []
    for num_frames in range(100):
        o, r, tm, tc, info = env.step([0, 1])
        frame = o['image']
        frame = frame[..., 0]  # Original return frame is [1200, 1920, 3, 1] (float), so remove last dim.
        # frame = 1 - frame
        # frame *= 255
        # frame = frame.astype(np.uint8)
        # print(f"Finish {num_frames + 1} frames")
        # plt.imshow(frame)
        # plt.show()
        # print(
        #     "Finish {}/100 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
        #         num_frames + 1,
        #         time.time() - start, (num_frames + 1) / (time.time() - start)
        #     )
        # )
    env.close()



# Author: Epihaius
# Date: 2019-06-21
#
# This module contains a class to generate cone primitives.

from .base import *


class ConeMaker(ModelMaker):
    @property
    def bottom_center(self):
        return self._bottom_center

    @bottom_center.setter
    def bottom_center(self, pos):
        self._bottom_center = pos

    @property
    def top_center(self):
        return self._top_center

    @top_center.setter
    def top_center(self, pos):
        self._top_center = pos

    @property
    def bottom_radius(self):
        return self._bottom_radius

    @bottom_radius.setter
    def bottom_radius(self, radius):
        self._bottom_radius = radius

    @property
    def top_radius(self):
        return self._top_radius

    @top_radius.setter
    def top_radius(self, radius):
        self._top_radius = radius

    @property
    def smooth(self):
        return self._smooth

    @smooth.setter
    def smooth(self, smooth):
        self._smooth = smooth

    @property
    def slice(self):
        return self._slice

    @slice.setter
    def slice(self, angle):
        self._slice = angle

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, angle):
        self._rotation = angle

    @property
    def bottom_thickness(self):
        return self._bottom_thickness

    @bottom_thickness.setter
    def bottom_thickness(self, thickness):
        self._bottom_thickness = thickness

    @property
    def top_thickness(self):
        return self._top_thickness

    @top_thickness.setter
    def top_thickness(self, thickness):
        self._top_thickness = thickness

    def __init__(
        self,
        bottom_center=None,
        top_center=None,
        bottom_radius=1.,
        top_radius=0.,
        segments=None,
        smooth=True,
        slice=0.,
        rotation=0.,
        bottom_thickness=None,
        top_thickness=None,
        inverted=False,
        vertex_color=None,
        has_uvs=True,
        tex_units=None,
        tex_offset=None,
        tex_rotation=None,
        tex_scale=None
    ):
        """
        This class generates cone model primitives with the given parameters:

            bottom_center (sequence or None):
                the position of the bottom center in object space;
                default is at the origin (the object-space (0., 0., 0.) point);
            top_center (sequence or None):
                the position of the top center in object space;
                default is at (0., 0., 1.);
            bottom_radius (float):
                the bottom radius of the cone; cannot be negative;
                default = 1.;
            top_radius (float):
                the top radius of the cone; cannot be negative;
                default = 0.;
            segments (dict of ints):
                the number of surface subdivisions:
                    "circular":
                        subdivisions of the mantle along a circular cross-section;
                        minimum = 3, default = 20;
                    "axial":
                        subdivisions of the mantle along the axis of rotation;
                        minimum = 1, default = 1;
                    "bottom_cap":
                        radial subdivisions of the bottom cap;
                        minimum = 0 (no cap), default = 1;
                    "top_cap":
                        radial subdivisions of the top cap;
                        minimum = 0 (no cap), default = 1;
                    "slice_caps_radial":
                        subdivisions of both slice caps, along the radius;
                        minimum = 0 (no caps), default = 1;
                    "slice_caps_axial":
                        subdivisions of both slice caps, along the axis of rotation;
                        minimum = 0 (no caps), default = 1;
            smooth (bool):
                whether the surface of the mantle should appear smooth or faceted;
                default is True;
            slice (float):
                the angle of the pie slice removed from the cone, in degrees;
                must be in [0., 360.] range; starts at the positive X-axis;
                default = 0. (no slice);
            rotation (float):
                the angle at which the cone is rotated about its local axis,
                in degrees;
                default = 0.;
            bottom_thickness (float or None):
                bottom radial offset of inner cone;
                creates an inner cone if smaller than bottom_radius;
                must be in [0., bottom_radius] range;
                default = None (no inner cone if top_thickness is also None);
            top_thickness (float or None):
                top radial offset of inner cone;
                creates an inner cone if smaller than top_radius;
                must be in [0., top_radius] range;
                default = None (no inner cone if bottom_thickness is also None).

        The parameters common to all primitive types are documented in the
        ModelMaker base class.

        The surfaces of this primitive type can be referred to in the relevant
        parameters and properties as follows:
            "main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap",
            "inner_main".

        Ranges of vertex indices of each surface can be retrieved through the
        vertex_ranges property, as a dict of (start_index, end_index) tuples
        (empty tuple if the surface was not created), with end_index not
        included in the range.

        """

        surface_ids = ("main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap", "inner_main")

        ModelMaker.__init__(
            self, segments, inverted, vertex_color, has_uvs, tex_units, tex_offset, tex_rotation, tex_scale, surface_ids
        )

        self._bottom_center = bottom_center
        self._top_center = top_center
        self._bottom_radius = bottom_radius
        self._top_radius = top_radius
        self._smooth = smooth
        self._slice = slice
        self._rotation = rotation
        self._bottom_thickness = bottom_thickness
        self._top_thickness = top_thickness

    def reset(self):

        ModelMaker.reset(self)

        self._bottom_center = None
        self._top_center = None
        self._bottom_radius = 1.
        self._top_radius = 0.
        self._smooth = True
        self._slice = 0.
        self._rotation = 0.
        self._bottom_thickness = None
        self._top_thickness = None

    def __transform_vertices(self, vertex_data, axis_vec, bottom_center):

        mat = Mat4(Mat4.ident_mat())

        if self._rotation:
            mat *= Mat4.rotate_mat(self._rotation, Vec3.up())

        if axis_vec.normalize():

            cross_vec = axis_vec.cross(Vec3.up())
            ref_vec = cross_vec if cross_vec.normalize() else Vec3.right()
            angle = Vec3.up().signed_angle_deg(axis_vec, ref_vec)

            if angle:
                mat *= Mat4.rotate_mat(angle, ref_vec)

        x, y, z = bottom_center

        if x or y or z:
            mat *= Mat4.translate_mat(x, y, z)

        vertex_data.transform_vertices(mat)

    def generate(self):

        bottom_center = (0., 0., 0.) if self._bottom_center is None else self._bottom_center
        top_center = (0., 0., 1.) if self._top_center is None else self._top_center
        axis_vec = Point3(*top_center) - Point3(*bottom_center)
        height = axis_vec.length()
        bottom_radius = max(0., self._bottom_radius)
        top_radius = max(0., self._top_radius)

        if bottom_radius == top_radius == 0.:
            bottom_radius = .001

        segs = {} if self._segments is None else self._segments
        segs_c = max(3, segs.get("circular", 20))
        segs_a = max(1, segs.get("axial", 1))
        segs_bc = max(0, segs.get("bottom_cap", 1) if bottom_radius else 0)
        segs_tc = max(0, segs.get("top_cap", 1) if top_radius else 0)
        segs_sc_r = max(0, segs.get("slice_caps_radial", 1))
        segs_sc_a = max(0, segs.get("slice_caps_axial", 1))
        smooth = self._smooth
        slice = max(0., min(360., self._slice))
        slice_radians = pi * slice / 180.
        delta_angle = pi * ((360. - slice) / 180.) / segs_c
        delta_radius = top_radius - bottom_radius
        bottom_thickness = bottom_radius if self._bottom_thickness is None \
                           else max(0., min(bottom_radius, self._bottom_thickness))
        inner_bottom_radius = bottom_radius - bottom_thickness
        top_thickness = top_radius if self._top_thickness is None \
                           else max(0., min(top_radius, self._top_thickness))
        inner_top_radius = top_radius - top_thickness
        inverted = self._inverted
        has_uvs = self._has_uvs
        tex_units = self._tex_units
        tex_offset = self._tex_offset
        tex_rotation = self._tex_rotation
        tex_scale = self._tex_scale
        self._vert_ranges = vert_ranges = {
            "main": (),
            "bottom_cap": (),
            "top_cap": (),
            "slice_start_cap": (),
            "slice_end_cap": (),
            "inner_main": ()
        }
        stride = 8 if has_uvs else 6  # number of floats on each vertex data row
        values = array.array("f", [])
        indices = array.array("I", [])
        verts = []

        normal = (0., 0., 1. if inverted else -1.)

        if has_uvs:

            if tex_units and "bottom_cap" in tex_units:
                tex_size = tex_units["bottom_cap"]
            else:
                tex_size = None

            mat = self._get_tex_xform("bottom_cap")

        if segs_bc and not inner_bottom_radius:

            # Define the bottom cap triangle vertices

            u = v = .5

            if has_uvs and mat:
                u, v = mat.xform_point(Point2(u, v))

            vert = {"pos": (0., 0., 0.), "normal": normal, "uv": (u, v)}
            verts.append(vert)

            radius = bottom_radius / segs_bc

            for i in range(segs_c + 1):

                angle = delta_angle * i + (0. if inverted else slice_radians)
                c = cos(angle)
                s = sin(angle) * (-1. if inverted else 1.)
                x = radius * c
                y = radius * s

                if has_uvs:

                    u = .5 + .5 * c / segs_bc
                    v = .5 + .5 * s * (1. if inverted else -1.) / segs_bc

                    if tex_size:
                        u = (u - .5) * 2. * bottom_radius / tex_size[0] + .5
                        v = (v - .5) * 2. * bottom_radius / tex_size[1] + .5

                    if mat:
                        u, v = mat.xform_point(Point2(u, v))

                else:

                    u = v = 0.

                vert = {"pos": (x, y, 0.), "normal": normal, "uv": (u, v)}
                verts.append(vert)

            # Define the vertex order of the bottom cap triangles

            for i in range(1, segs_c + 1):
                indices.extend((0, i + 1, i))

        if segs_bc and bottom_thickness:

            # Define the bottom cap quad vertices

            n = 0 if inner_bottom_radius else 1

            for i in range(n, segs_bc + 1 - n):

                radius = inner_bottom_radius + bottom_thickness * (i + n) / segs_bc

                for j in range(segs_c + 1):

                    angle = delta_angle * j + (0. if inverted else slice_radians)
                    c = cos(angle)
                    s = sin(angle) * (-1. if inverted else 1.)
                    x = radius * c
                    y = radius * s

                    if has_uvs:

                        r = radius / bottom_radius
                        u = .5 + .5 * c * r
                        v = .5 + .5 * s * (1. if inverted else -1.) * r

                        if tex_size:
                            u = (u - .5) * 2. * bottom_radius / tex_size[0] + .5
                            v = (v - .5) * 2. * bottom_radius / tex_size[1] + .5

                        if mat:
                            u, v = mat.xform_point(Point2(u, v))

                    else:

                        u = v = 0.

                    vert = {"pos": (x, y, 0.), "normal": normal, "uv": (u, v)}
                    verts.append(vert)

            # Define the vertex order of the bottom cap quads

            index_offset = segs_c + 1 if inner_bottom_radius else 1

            for i in range(0 if inner_bottom_radius else 1, segs_bc):
                for j in range(segs_c):
                    vi1 = index_offset + i * (segs_c + 1) + j
                    vi2 = vi1 - segs_c - 1
                    vi3 = vi2 + 1
                    vi4 = vi1 + 1
                    indices.extend((vi1, vi2, vi3))
                    indices.extend((vi1, vi3, vi4))

            vert_ranges["bottom_cap"] = (0, len(verts))

        vertex_count = len(verts)

        if has_uvs:

            if tex_units and "main" in tex_units:
                tex_size = tex_units["main"]
                max_radius = max(bottom_radius, top_radius)
                arc = (2 * pi - slice_radians) * max_radius
            else:
                tex_size = None

            mat = self._get_tex_xform("main")

        # Define the mantle quad vertices

        for i in range(segs_a + 1):

            radius = bottom_radius + delta_radius * i / segs_a
            z = height * i / segs_a

            if has_uvs:
                v = i / segs_a
                if tex_size:
                    v *= height / tex_size[1]
                v_start = v
            else:
                v = 0.

            for j in range(segs_c + 1):

                angle = delta_angle * j + (0. if inverted else slice_radians)
                x = radius * cos(angle)
                y = radius * sin(angle) * (-1. if inverted else 1.)

                if smooth:
                    normal = Vec3(x, y, -radius * delta_radius / height).normalized()
                    normal *= -1. if inverted else 1.

                if has_uvs:
                    u = j / segs_c
                    if tex_size:
                        u *= arc / tex_size[0]
                    if mat:
                        u, v = mat.xform_point(Point2(u, v_start))
                else:
                    u = 0.

                vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

                if not smooth and 0 < j < segs_c:
                    verts.append(vert.copy())

        # Define the vertex order of the mantle quads

        n = segs_c + 1 if smooth else segs_c * 2
        f = 1 if smooth else 2

        for i in range(1, segs_a + 1):

            for j in range(0, segs_c * f, f):

                vi1 = vertex_count + i * n + j
                vi2 = vi1 - n
                vi3 = vi2 + 1
                vi4 = vi1 + 1
                indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi2, vi3))
                indices.extend((vi2, vi3, vi4) if inverted else (vi1, vi3, vi4))

                if not smooth:
                    self._make_flat_shaded((vi1, vi2, vi3, vi4), verts)

        vert_ranges["main"] = (vertex_count, len(verts))

        vertex_count = len(verts)

        normal = (0., 0., -1. if inverted else 1.)

        if has_uvs:

            if tex_units and "top_cap" in tex_units:
                tex_size = tex_units["top_cap"]
            else:
                tex_size = None

            mat = self._get_tex_xform("top_cap")

        if segs_tc and not inner_top_radius:

            index_offset = vertex_count

            # Define the top cap triangle vertices

            u = v = .5

            if has_uvs and mat:
                u, v = mat.xform_point(Point2(u, v))

            vert = {"pos": (0., 0., height), "normal": normal, "uv": (u, v)}
            verts.append(vert)

            radius = top_radius / segs_tc

            for i in range(segs_c + 1):

                angle = delta_angle * i + (0. if inverted else slice_radians)
                c = cos(angle)
                s = sin(angle) * (-1. if inverted else 1.)
                x = radius * c
                y = radius * s

                if has_uvs:

                    u = .5 + .5 * c / segs_tc
                    v = .5 + .5 * s * (-1. if inverted else 1.) / segs_tc

                    if tex_size:
                        u = (u - .5) * 2. * top_radius / tex_size[0] + .5
                        v = (v - .5) * 2. * top_radius / tex_size[1] + .5

                    if mat:
                        u, v = mat.xform_point(Point2(u, v))

                else:

                    u = v = 0.

                vert = {"pos": (x, y, height), "normal": normal, "uv": (u, v)}
                verts.append(vert)

            # Define the vertex order of the top cap triangles

            for i in range(vertex_count + 1, vertex_count + 1 + segs_c):
                indices.extend((vertex_count, i, i + 1))

        # Define the top cap quad vertices

        if segs_tc and top_thickness:

            n = 0 if inner_top_radius else 1

            for i in range(n, segs_tc + 1 - n):

                radius = inner_top_radius + top_thickness * (i + n) / segs_tc

                for j in range(segs_c + 1):

                    angle = delta_angle * j + (0. if inverted else slice_radians)
                    c = cos(angle)
                    s = sin(angle) * (-1. if inverted else 1.)
                    x = radius * c
                    y = radius * s

                    if has_uvs:

                        r = radius / top_radius
                        u = .5 + .5 * c * r
                        v = .5 + .5 * s * (-1. if inverted else 1.) * r

                        if tex_size:
                            u = (u - .5) * 2. * top_radius / tex_size[0] + .5
                            v = (v - .5) * 2. * top_radius / tex_size[1] + .5

                        if mat:
                            u, v = mat.xform_point(Point2(u, v))

                    else:

                        u = v = 0.

                    vert = {"pos": (x, y, height), "normal": normal, "uv": (u, v)}
                    verts.append(vert)

            # Define the vertex order of the top cap quads

            index_offset = vertex_count + (segs_c + 1 if inner_top_radius else 1)

            for i in range(0 if inner_top_radius else 1, segs_tc):
                for j in range(segs_c):
                    vi1 = index_offset + i * (segs_c + 1) + j
                    vi2 = vi1 - segs_c - 1
                    vi3 = vi2 + 1
                    vi4 = vi1 + 1
                    indices.extend((vi1, vi3, vi2))
                    indices.extend((vi1, vi4, vi3))

        vert_ranges["top_cap"] = (vertex_count, len(verts))

        if segs_sc_r and segs_sc_a and slice and (bottom_thickness or top_thickness):

            # Define the slice cap vertices

            index_offset = len(verts)
            max_radius = max(bottom_radius, top_radius)
            delta_inner_radius = inner_top_radius - inner_bottom_radius

            for cap_id in ("start", "end"):

                vertex_count = len(verts)

                if cap_id == "start":
                    normal = (0., -1. if inverted else 1., 0.)
                else:
                    angle = delta_angle * segs_c
                    c = cos(angle)
                    s = -sin(angle)
                    normal = Vec3(s, -c, 0.) * (-1. if inverted else 1.)

                if has_uvs:

                    cap_name = "slice_{}_cap".format(cap_id)
                    if tex_units and cap_name in tex_units:
                        tex_size = tex_units[cap_name]
                    else:
                        tex_size = None

                    mat = self._get_tex_xform(cap_name)

                for i in range(segs_sc_a + 1):

                    # Define the vertices of the slice cap quad

                    f = i / segs_sc_a
                    radius = bottom_radius + delta_radius * f
                    inner_radius = inner_bottom_radius + delta_inner_radius * f
                    z = height * f

                    if has_uvs:
                        v = f
                        if tex_size:
                            v *= height / tex_size[1]
                        v_start = v
                    else:
                        v = 0.

                    for j in range(segs_sc_r + 1):

                        r = inner_radius + (radius - inner_radius) * j / segs_sc_r

                        if cap_id == "start":
                            pos = (r, 0., z)
                        else:
                            pos = (r * c, r * s, z)

                        if has_uvs:

                            if cap_id == "start":
                                u = .5 + .5 * r / max_radius * (1. if inverted else -1.)
                            else:
                                u = .5 - .5 * r / max_radius * (1. if inverted else -1.)

                            if tex_size:
                                u = (u - .5) * 2. * max_radius / tex_size[0] + .5

                            if mat:
                                u, v = mat.xform_point(Point2(u, v_start))

                        else:

                            u = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                for i in range(segs_sc_a):

                    # Define the vertex order of the slice cap quad

                    for j in range(segs_sc_r):

                        vi1 = index_offset + j
                        vi2 = vi1 + segs_sc_r + 1
                        vi3 = vi1 + 1
                        vi4 = vi2 + 1

                        if cap_id == "start":
                            indices.extend((vi1, vi3, vi2) if inverted else (vi1, vi2, vi3))
                            indices.extend((vi2, vi3, vi4) if inverted else (vi2, vi4, vi3))
                        else:
                            indices.extend((vi1, vi2, vi3) if inverted else (vi1, vi3, vi2))
                            indices.extend((vi2, vi4, vi3) if inverted else (vi2, vi3, vi4))

                    index_offset += segs_sc_r + 1

                index_offset += segs_sc_r + 1

                surface_name = "slice_{}_cap".format(cap_id)
                vert_ranges[surface_name] = (vertex_count, len(verts))

        for vert in verts:

            values.extend(vert["pos"])
            values.extend(vert["normal"])

            if has_uvs:
                values.extend(vert["uv"])

        # Create the geometry structures

        if inner_bottom_radius or inner_top_radius:

            # If a thickness is given, an inner cone needs to be created to close
            # the surface of the model; its parameters are derived from those of
            # the outer cone and adjusted to make both cones fit together.

            segs = {
                "circular": segs_c,
                "axial": segs_a,
                "bottom_cap": 0,
                "top_cap": 0,
                "slice_caps_radial": 0,
                "slice_caps_axial": 0
            }
            inner_tex_units = {} if tex_units else None

            if tex_units and "inner_main" in tex_units:
                inner_tex_units["main"] = tex_units["inner_main"]

            inner_tex_offset = {} if tex_offset else None

            if tex_offset and "inner_main" in tex_offset:
                inner_tex_offset["main"] = tex_offset["inner_main"]

            inner_tex_rot = {} if tex_rotation else None

            if tex_rotation and "inner_main" in tex_rotation:
                inner_tex_rot["main"] = tex_rotation["inner_main"]

            inner_tex_scale = {} if tex_scale else None

            if tex_scale and "inner_main" in tex_scale:
                inner_tex_scale["main"] = tex_scale["inner_main"]

            model_maker = ConeMaker(
                None, (0., 0., height),
                inner_bottom_radius,
                inner_top_radius,
                segs,
                smooth,
                slice,
                inverted=not inverted,
                has_uvs=has_uvs,
                tex_units=inner_tex_units,
                tex_offset=inner_tex_offset,
                tex_rotation=inner_tex_rot,
                tex_scale=inner_tex_scale
            )
            node = model_maker.generate()

            # Extend the geometry of the inner cone with the data of the outer cone

            geom = node.modify_geom(0)
            vertex_data = geom.modify_vertex_data()
            old_vert_count = vertex_data.get_num_rows()
            old_size = old_vert_count * stride
            vertex_data.set_num_rows(old_vert_count + len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[old_size:] = values
            self.__transform_vertices(vertex_data, axis_vec, bottom_center)

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)
                geom.set_vertex_data(vertex_data)

            tris_prim = geom.modify_primitive(0)
            old_row_count = tris_prim.get_num_vertices()
            new_row_count = old_row_count + len(indices)

            if new_row_count < 2**16:
                # make the array compatible with the default index format of the
                # GeomPrimitive (16-bit) if the number of vertices allows it...
                indices = array.array("H", indices)
            else:
                # ...or force the GeomPrimitive to accept more indices by setting
                # its index format to 32-bit
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.set_num_rows(new_row_count)
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[old_row_count:] = indices
            tris_prim.offset_vertices(old_vert_count, old_row_count, new_row_count)

            inner_range = model_maker.vertex_ranges["main"]

            if inner_range:
                vert_ranges["inner_main"] = inner_range

            for surface_name in ("main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap"):

                vert_range = vert_ranges[surface_name]

                if vert_range:
                    start, end = vert_range
                    start += old_vert_count
                    end += old_vert_count
                    vert_ranges[surface_name] = (start, end)

        else:

            if has_uvs:
                vertex_format = GeomVertexFormat.get_v3n3t2()
            else:
                vertex_format = GeomVertexFormat.get_v3n3()

            vertex_data = GeomVertexData("cone_data", vertex_format, Geom.UH_static)
            vertex_data.unclean_set_num_rows(len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[:] = values
            self.__transform_vertices(vertex_data, axis_vec, bottom_center)

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)

            tris_prim = GeomTriangles(Geom.UH_static)

            if len(indices) < 2**16:
                indices = array.array("H", indices)
            else:
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.unclean_set_num_rows(len(indices))
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[:] = indices

            geom = Geom(vertex_data)
            geom.add_primitive(tris_prim)
            node = GeomNode("cone")
            node.add_geom(geom)

        return node



import cv2
from panda3d.core import Shader, RenderState, ShaderAttrib, GeoMipTerrain, PNMImage, Texture

from metadrive.component.sensors.base_camera import BaseCamera
from metadrive.constants import CamMask
from metadrive.constants import RENDER_MODE_NONE
from metadrive.engine.asset_loader import AssetLoader


class DepthCamera(BaseCamera):
    # shape(dim_1, dim_2)
    CAM_MASK = CamMask.DepthCam

    GROUND_HEIGHT = -0.5
    VIEW_GROUND = False
    GROUND = None
    GROUND_MODEL = None

    def __init__(self, engine, width, height, *, cuda=False):
        self.BUFFER_W, self.BUFFER_H = width, height
        self.VIEW_GROUND = True  # default true
        super(DepthCamera, self).__init__(engine, False, cuda)
        cam = self.get_cam()
        lens = self.get_lens()

        # cam.lookAt(0, 2.4, 1.3)
        cam.lookAt(0, 10.4, 1.6)

        lens.setFov(60)
        # lens.setAspectRatio(2.0)
        if self.engine.mode == RENDER_MODE_NONE or not AssetLoader.initialized():
            return
        # add shader for it
        # if get_global_config()["headless_machine_render"]:
        #     vert_path = AssetLoader.file_path("shaders", "depth_cam_gles.vert.glsl")
        #     frag_path = AssetLoader.file_path("shaders", "depth_cam_gles.frag.glsl")
        # else:
        from metadrive.utils import is_mac
        if is_mac():
            vert_path = AssetLoader.file_path("shaders", "depth_cam_mac.vert.glsl")
            frag_path = AssetLoader.file_path("shaders", "depth_cam_mac.frag.glsl")
        else:
            vert_path = AssetLoader.file_path("shaders", "depth_cam.vert.glsl")
            frag_path = AssetLoader.file_path("shaders", "depth_cam.frag.glsl")
        custom_shader = Shader.load(Shader.SL_GLSL, vertex=vert_path, fragment=frag_path)
        cam.node().setInitialState(RenderState.make(ShaderAttrib.make(custom_shader, 1)))

        if self.VIEW_GROUND:
            ground = PNMImage(257, 257, 4)
            ground.fill(1., 1., 1.)

            self.GROUND = GeoMipTerrain("mySimpleTerrain")
            self.GROUND.setHeightfield(ground)
            self.GROUND.setAutoFlatten(GeoMipTerrain.AFMStrong)
            # terrain.setBruteforce(True)
            # # Since the terrain is a texture, shader will not calculate the depth information, we add a moving terrain
            # # model to enable the depth information of terrain
            self.GROUND_MODEL = self.GROUND.getRoot()
            self.GROUND_MODEL.setPos(-128, -128, self.GROUND_HEIGHT)
            self.GROUND_MODEL.reparentTo(self.engine.render)
            self.GROUND_MODEL.hide(CamMask.AllOn)
            self.GROUND_MODEL.show(CamMask.DepthCam)
            self.GROUND.generate()

    def track(self, base_object):
        if self.VIEW_GROUND:
            pos = base_object.origin.getPos()
            self.GROUND_MODEL.setPos(pos[0], pos[1], self.GROUND_HEIGHT)
            # self.GROUND_MODEL.setP(-base_object.origin.getR())
            # self.GROUND_MODEL.setR(-base_object.origin.getR())
        return super(DepthCamera, self).track(base_object)

    def get_image(self, base_object):
        self.origin.reparentTo(base_object.origin)
        img = super(DepthCamera, self).get_rgb_array_cpu()
        self.track(self.attached_object)
        return img

    def save_image(self, base_object, name="debug.png"):
        img = self.get_image(base_object)
        cv2.imwrite(name, img)



import numpy as np
from panda3d.core import Vec3
from metadrive.utils.math import Vector
from metadrive.utils.math import wrap_to_pi

#
# Now all coordinates are the same and are all in right-handed
# MetaDrive (right handed):
#                     ^ x
#                     |
#                     |
#         y<----------|
#               EgoCar/Object/Pedestrian/Cyclist and so on
#
# Panda3d (right handed):
#                     ^ x
#                     |
#                     |
#         y <---------|
#                world origin
# Note: Camera is facing to y-axis:
#                         ^ x
#                         |
#                         |
#                  | \    |
#         y <------|  |---|
#                  | /    |
#                camera origin

# Note: the models loaded in Panda3d are facing to y axis, and thus -90' is required to make it face to x axis
# These APIs are still available for compatibility, but doesn't apply any operation to the input vector/heading


def panda_vector(position, z=0.0) -> Vec3:
    """
    Give a 2d or 3d position in MetaDrive, transform it to Panda3d world.
    If the position is a 2d array, height will be determined by the value of z.
    if the position is a 3d array, the value of z will be invalid.
    :param position: 2d or 3d position
    :param z: the height of the position, when position is a 2d array
    :return: position represented in Vec3
    """
    if len(position) == 3:
        z = position[2]
    return Vec3(position[0], position[1], z)


def metadrive_vector(position):
    """
    Transform the position in Panda3d to MetaDrive world
    :param position: Vec3, position in Panda3d
    :return: 2d position
    """
    # return np.array([position[0], -position[1]])
    # return position[0], -position[1]
    return Vector([position[0], position[1]])


def panda_heading(heading: float) -> float:
    """
    Transform the heading in MetaDrive to Panda3d
    :param heading: float, heading in MetaDrive (degree)
    :return: heading (degree)
    """
    # return -heading
    return heading


def metadrive_heading(heading: float) -> float:
    """
    Transform the heading in Panda3d to MetaDrive
    :param heading: float, heading in panda3d (degree)
    :return: heading (degree)
    """
    # return -heading
    return heading


def nuplan_to_metadrive_vector(vector, nuplan_center=(0, 0)):
    "All vec in nuplan should be centered in (0,0) to avoid numerical explosion"
    vector = np.array(vector)
    # if len(vector.shape) == 1:
    #     vector[1] *= -1
    # else:
    #     vector[:, 1] *= -1
    vector -= np.asarray(nuplan_center)
    return vector


def metadrive_to_nuplan_vector(vector, nuplan_center=(0, 0)):
    "All vec in nuplan should be centered in (0,0) to avoid numerical explosion"
    vector = np.array(vector)
    vector += nuplan_center
    # if len(vector.shape) == 1:
    #     vector[1] *= -1
    # else:
    #     vector[:, 1] *= -1
    return vector


def nuplan_to_metadrive_heading(heading):
    # return -wrap_to_pi(heading)
    return wrap_to_pi(heading)


def metadrive_to_nuplan_heading(heading):
    # return -wrap_to_pi(heading)
    return wrap_to_pi(heading)
